import os
from glob import glob

from ievv_opensource.utils.ievvbuildstatic import pluginbase
from ievv_opensource.utils.ievvbuildstatic.filepath import \
    AbstractDjangoAppPath
from ievv_opensource.utils.ievvbuildstatic.watcher import ProcessWatchConfig
from ievv_opensource.utils.shellcommandmixin import (ShellCommandError,
                                                     ShellCommandMixin)

from .gzip_compress_mixin import GzipCompressMixin


class Plugin(pluginbase.Plugin, ShellCommandMixin, GzipCompressMixin):
    """
    Webpack builder plugin.

    Examples:

        Simple example::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.npmrun_jsbuild.Plugin(),
                    ]
                )
            )

        Webpack example::

            Install webpack:

            $ yarn add webpack

            Add the following to your package.json:

                {
                    ...
                    "scripts": {
                        ...
                        "jsbuild": "webpack --config webpack.config.js",
                        "jsbuild-production": "webpack --config webpack.config.js -p"
                        ...
                    }
                    ...
                }

            Create a webpack.config.js with something like this:

                let path = require('path');

                const isProduction = process.env.IEVV_BUILDSTATIC_MODE == 'production';
                const appconfig = require("./ievv_buildstatic.appconfig.json");
                console.log(isProduction);
                console.log(appconfig);

                let webpackConfig = {
                    entry: './scripts/javascript/ievv_jsbase/ievv_jsbase_core.js',
                    output: {
                        filename: 'ievv_jsbase_core.js',
                        path: path.resolve(appconfig.destinationfolder, 'scripts')
                    },
                    module: {
                        loaders: [
                            {
                                test: /.jsx?$/,
                                loader: 'babel-loader',
                                // exclude: /node_modules/
                                include: [
                                    path.resolve(__dirname, "scripts/javascript/ievv_jsbase"),
                                ]
                            }
                        ]
                    }
                };

                if(isProduction) {
                    webpackConfig.devtool = 'source-map';
                } else {
                    webpackConfig.devtool = 'cheap-module-eval-source-map';
                    webpackConfig.output.pathinfo = true;
                }

                module.exports = webpackConfig;

    """
    name = 'npmrun_jsbuild'
    default_group = 'js'

    def __init__(self,
                 extra_import_paths=None,
                 extra_import_aliases=None,
                 gzip=False, gzip_compresslevel=9,
                 **kwargs):
        """
        
        Args:
            extra_import_paths (list, optional): List of extra javascript import paths. Defaults to None.
            extra_import_aliases (list, optional): Mapping of extra import aliases. Defaults to None.
            gzip (bool, optional): Make a .js.gz version of the file in --production mode. This is added in
                addition to the .js files (same filename, just with .gz at the end).
            gzip_compresslevel (int, optional): Gzip compression level.
                A number between 0 and 9. Higher is better compression,
                but slower to compress. Defaults to 9.
        """
        super(Plugin, self).__init__(**kwargs)
        self.extra_import_paths = extra_import_paths or []
        self.extra_import_aliases = extra_import_aliases or {}
        self.gzip = gzip
        self.gzip_compresslevel = gzip_compresslevel

    # @property
    # def destinationfolder(self):
    #     return self.app.get_destination_path('scripts')

    def gzip_compression_enabled(self):
        return self.gzip and self.app.apps.is_in_production_mode()

    def gzip_compresslevel(self):
        return self.gzip_compresslevel

    def get_filepaths_to_gzip(self):
        return glob(os.path.join(self.app.get_destination_path(), '**/*.js'), recursive=True)

    def get_default_import_paths(self):
        return []

    def get_import_paths(self):
        return self.get_default_import_paths() + self.extra_import_paths

    def _get_import_paths_as_strlist(self):
        import_paths = []
        for path in self.get_import_paths():
            if isinstance(path, AbstractDjangoAppPath):
                path = path.abspath
            import_paths.append(path)
        return import_paths

    def _get_extra_import_aliases(self):
        import_aliases = {}
        for alias, path in self.extra_import_aliases.items():
            if isinstance(path, AbstractDjangoAppPath):
                path = path.abspath
            import_aliases[alias] = path
        return import_aliases

    def install(self):
        self.app.add_pluginconfig_to_json_config(
            plugin_name=self.name,
            config_dict={
                'import_paths': self._get_import_paths_as_strlist(),
                'extra_import_aliases': self._get_extra_import_aliases()
            }
        )

    def get_npm_script(self):
        if self.app.apps.is_in_production_mode():
            return 'jsbuild-production'
        else:
            return 'jsbuild'

    def get_npm_watch_script(self):
        return 'jsbuild-watch'

    def run(self):
        npm_script = self.get_npm_script()
        about = '"npm run {npm_script}" for {appname!r}'.format(
            npm_script=npm_script,
            appname=self.app.appname
        )
        self.get_logger().command_start('Running {about}'.format(
            about=about))
        try:
            self.run_shell_command('npm',
                                   args=['run', npm_script],
                                   _cwd=self.app.get_source_path())
        except ShellCommandError:
            self.get_logger().command_error('{} FAILED!'.format(about))
        else:
            self.gzip_compress_files()
            self.get_logger().command_success('{} succeeded :)'.format(about))

    def __str__(self):
        return '{}({})'.format(super(Plugin, self).__str__(), self.sourcefolder)

    def run_watcher_process(self):
        about = '"npm run {scriptname}" for {appname!r}'.format(
            scriptname=self.get_npm_watch_script(),
            appname=self.app.appname)
        self.get_logger().info(
            'Starting watcher process: {about}.'.format(about=about))
        try:
            self.app.get_installer('npm').run_packagejson_script(script=self.get_npm_watch_script())
        except ShellCommandError:
            self.get_logger().command_error('{} FAILED!'.format(about))
        except KeyboardInterrupt:
            pass

    def watch(self):
        if self.app.get_installer('npm').has_npm_script(self.get_npm_watch_script()):
            return ProcessWatchConfig(
                plugin=self)
        else:
            return None



from django import test

from ievv_opensource.ievv_sms.backends import debug_dbstore
from ievv_opensource.ievv_sms.models import DebugSmsMessage


class TestBackend(test.TestCase):
    def test_sms_record_is_created_in_db(self):
        debug = debug_dbstore.Backend(message="Hi there!", phone_number="1234567", send_as='7654321')
        debug.send()
        self.assertEqual(
            DebugSmsMessage.objects.first().message,
            "Hi there!")
        self.assertEqual(
            DebugSmsMessage.objects.first().phone_number,
            "1234567")
        self.assertEqual(
            DebugSmsMessage.objects.first().send_as,
            "7654321")

    def test_sms_record_send_as_default_blank(self):
        debug = debug_dbstore.Backend(message="Hi there!", phone_number="1234567")
        debug.send()
        self.assertEqual(
            DebugSmsMessage.objects.first().send_as,
            '')



import os
import sys


def is_in_virtualenv():
    """
    Returns ``True`` if we are in a virtualenv.
    """
    return hasattr(sys, 'real_prefix') or hasattr(sys, 'base_prefix')


def get_virtualenv_directory():
    """
    Get the root directory of the current virtualenv.

    Raises OSError if not in a virtualenv.
    """
    if is_in_virtualenv():
        return os.path.abspath(sys.prefix)
    else:
        raise OSError('Not in a virtualenv.')


def add_virtualenv_bin_directory_to_path():
    """
    Add :func:`get_virtualenv_directory` to ``os.environ['PATH']``.

    Why do we need this? This is used to work around limitations
    in how certain IDE's implement virtualenv support. They may
    add the virtualenv to PYTHONPATH, but to to PATH.
    """
    bin_directory = os.path.join(get_virtualenv_directory(), 'bin')
    if bin_directory not in os.environ['PATH'].split(os.pathsep):
        os.environ["PATH"] += os.pathsep + bin_directory



from pkg_resources import resource_string
import json
import sys

if sys.version_info.major == 2:  # pragma: no cover
    __version__ = json.loads(resource_string(__name__, 'version.json'))
else:  # pragma: no cover
    __version__ = json.loads(resource_string(__name__, 'version.json').decode('utf-8'))



from __future__ import unicode_literals

from django.conf.urls import url
from django.contrib import admin
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.urls import reverse
from django.utils.html import format_html

from ievv_opensource.ievv_developemail.models import DevelopEmail


@admin.register(DevelopEmail)
class DevelopEmailAdmin(admin.ModelAdmin):
    list_display_links = ['get_edit']
    list_display = [
        'get_subject',
        'created_datetime',
        'from_email',
        'to_emails',
        'id',
        'get_edit',
    ]

    search_fields = [
        'id',
        'from_email',
        'to_emails',
        'raw_message',
    ]

    list_filter = [
        'created_datetime',
        'from_email',
    ]

    def get_edit(self, obj):
        return 'Edit'
    get_edit.short_description = 'Edit'

    def get_subject(self, developemail):
        view_url = reverse('admin:developemail_as_html', args=(developemail.id,))
        return format_html(
            '<a href="{url}">{label}</a>',
            url=view_url,
            label=developemail.subject_with_fallback)
    get_subject.short_description = 'Subject'
    get_subject.admin_order_field = 'subject'

    def get_urls(self):
        urls = super(DevelopEmailAdmin, self).get_urls()
        my_urls = [
            url(r'^(\d+)/as_html/$',
                self.admin_site.admin_view(self.as_html_view),
                name='developemail_as_html'),
            url(r'^(\d+)/as_plain/$',
                self.admin_site.admin_view(self.as_plain_view),
                name='developemail_as_plain'),
            url(r'^(\d+)/as_raw/$',
                self.admin_site.admin_view(self.as_raw_view),
                name='developemail_as_raw'),
        ]
        return my_urls + urls

    def _as_mail_preview_view(self, request, developemail_id, mode):
        context = {
            'developemail': get_object_or_404(DevelopEmail, id=developemail_id),
            'mode': mode,
            'developemail_count': DevelopEmail.objects.count()
        }
        return TemplateResponse(request, 'ievv_developemail/admin/as-{}.django.html'.format(mode), context)

    def as_plain_view(self, request, developemail_id):
        return self._as_mail_preview_view(request=request, developemail_id=developemail_id,
                                          mode='plain')

    def as_html_view(self, request, developemail_id):
        return self._as_mail_preview_view(request=request, developemail_id=developemail_id,
                                          mode='html')

    def as_raw_view(self, request, developemail_id):
        return self._as_mail_preview_view(request=request, developemail_id=developemail_id,
                                          mode='raw')



import os
from django.conf import settings
from django.core import management
import shutil
from django.test import TestCase


class TestRemoveSorlCacheMedia(TestCase):
    def setUp(self):
        self.thumbnail_prefix = 'test_sorlcache_dir'
        self.cachedir = os.path.join(settings.MEDIA_ROOT, self.thumbnail_prefix)
        if os.path.exists(self.cachedir):
            # This is not executed unless you run the tests multiple times
            # so we exclude it from coverage.
            shutil.rmtree(self.cachedir)  # pragma: no cover

    def test_no_cache_directory(self):
        with self.settings(THUMBNAIL_PREFIX=self.thumbnail_prefix):
            management.call_command('ievvtasks_remove_sorl_cache_media')

    def test_remove_existing_directory(self):
        os.makedirs(self.cachedir)
        open(os.path.join(self.cachedir, 'test.txt'), 'wb').write(b'test')
        with self.settings(THUMBNAIL_PREFIX=self.thumbnail_prefix):
            management.call_command('ievvtasks_remove_sorl_cache_media')
        self.assertFalse(os.path.exists(self.cachedir))



import os

from django.conf import settings
from django.core import management

from ievv_opensource.ievvtasks_common.base_command import BaseIevvTasksCommand


class Command(BaseIevvTasksCommand):
    help = 'Run makemessages for the languages specified in the ' \
           'IEVVTASKS_MAKEMESSAGES_LANGUAGE_CODES setting.'

    def __makemessages(self, ignore, extensions, domain):
        management.call_command(
            'makemessages',
            locale=settings.IEVVTASKS_MAKEMESSAGES_LANGUAGE_CODES,
            ignore=ignore,
            extensions=extensions,
            domain=domain)

    def __build_python_translations(self, config):
        if not config.get('python'):
            self.stdout.write(
                'Skipping python translations due to config["python"] != True. Config: {!r}'.format(
                    config))
            return
        default_ignore = getattr(settings, 'IEVVTASKS_MAKEMESSAGES_IGNORE', [
            'static/*'
        ])
        ignore = config.get('python_ignore', default_ignore)
        extensions = getattr(settings, 'IEVVTASKS_MAKEMESSAGES_EXTENSIONS', [
            'py', 'html', 'txt'])
        self.__makemessages(ignore=ignore,
                            extensions=extensions,
                            domain='django')

    def __build_javascript_translations(self, config):
        if not config.get('javascript'):
            self.stdout.write(
                'Skipping javascript translations due to config["javascript"] != True. Config: {!r}'.format(
                    config))
            return
        default_ignore = getattr(settings, 'IEVVTASKS_MAKEMESSAGES_JAVASCRIPT_IGNORE', [
            'node_modules/*',
            'bower_components/*',
            'not_for_deploy/*',
        ])
        ignore = config.get('javascript_ignore', default_ignore)
        extensions = getattr(settings, 'IEVVTASKS_MAKEMESSAGES_JAVASCRIPT_EXTENSIONS', [
            'js', 'jsx'])
        self.__makemessages(ignore=ignore,
                            extensions=extensions,
                            domain='djangojs')

    def __run_pre_management_commands(self):
        management_commands = getattr(
            settings, 'IEVVTASKS_MAKEMESSAGES_PRE_MANAGEMENT_COMMANDS', [])
        self.run_management_commands(management_commands)

    def handle(self, *args, **options):
        self.__run_pre_management_commands()
        current_directory = os.getcwd()
        for config in getattr(settings, 'IEVVTASKS_MAKEMESSAGES_DIRECTORIES', [current_directory]):
            if not isinstance(config, dict):
                config = {
                    'javascript': True,
                    'python': True,
                    'directory': config
                }
            directory = os.path.abspath(config['directory'])
            self.stdout.write('Running makemessages for python files in {}'.format(directory))
            os.chdir(directory)
            self.__build_python_translations(config)
            if getattr(settings, 'IEVVTASKS_MAKEMESSAGES_BUILD_JAVASCRIPT_TRANSLATIONS', False):
                self.stdout.write('Running makemessages for javascript files in {}'.format(directory))
                self.__build_javascript_translations(config)
            os.chdir(current_directory)



import re

NORMALIZE_WHITESPACE_REGEX = re.compile(r'(\s|\\xa0)+')


def normalize_whitespace(text):
    """
    Normalize the whitespace in the provided text so that
    all space characters (spaces, newlines, tags, ...) is
    replaced by a single space.

    Args:
        text (str): The string to normalize.

    Returns:
        str: The normalized text.
    """
    return NORMALIZE_WHITESPACE_REGEX.sub(' ', text)



import os

from ievv_opensource.utils.ievvbuildstatic import pluginbase
from ievv_opensource.utils.ievvbuildstatic.filepath import AbstractDjangoAppPath
from ievv_opensource.utils.shellcommandmixin import ShellCommandError
from ievv_opensource.utils.shellcommandmixin import ShellCommandMixin


class Plugin(pluginbase.Plugin, ShellCommandMixin):
    """
    Browserify javascript bundler plugin.

    Examples:

        Simple example::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.browserify_jsbuild.Plugin(
                            sourcefile='app.js',
                            destinationfile='app.js',
                        ),
                    ]
                )
            )

        Custom source folder example::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.browserify_jsbuild.Plugin(
                            sourcefolder=os.path.join('scripts', 'javascript', 'api'),
                            sourcefile='api.js',
                            destinationfile='api.js',
                        ),
                    ]
                )
            )
    """
    name = 'browserify_jsbuild'
    default_group = 'js'

    def __init__(self, sourcefile, destinationfile,
                 sourcefolder=os.path.join('scripts', 'javascript'),
                 destinationfolder=os.path.join('scripts'),
                 extra_watchfolders=None,
                 extra_import_paths=None,
                 sourcemap=True,
                 **kwargs):
        """
        Parameters:
            sourcefile: The source file relative to ``sourcefolder``.
            destinationfile: Path to destination file relative to ``destinationfolder``.
            sourcefolder: The folder where ``sourcefiles`` is located relative to
                the source folder of the :class:`~ievv_opensource.utils.ievvbuild.config.App`.
                Defaults to ``scripts/javascript``.
            destinationfolder: The folder where ``destinationfile`` is located relative to
                the destination folder of the :class:`~ievv_opensource.utils.ievvbuild.config.App`.
                Defaults to ``scripts/``.
            extra_watchfolders: List of extra folders to watch for changes.
                Relative to the source folder of the
                :class:`~ievv_opensource.utils.ievvbuild.config.App`.
            **kwargs: Kwargs for :class:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin`.
        """
        super(Plugin, self).__init__(**kwargs)
        self.sourcefile = sourcefile
        self.destinationfile = destinationfile
        self.destinationfolder = destinationfolder
        self.sourcefolder = sourcefolder
        self.extra_watchfolders = extra_watchfolders or []
        self.sourcemap = sourcemap
        self.extra_import_paths = extra_import_paths or []

    def get_sourcefile_path(self):
        return self.app.get_source_path(self.sourcefolder, self.sourcefile)

    def get_destinationfile_path(self):
        return self.app.get_destination_path(self.destinationfolder, self.destinationfile)

    def get_default_import_paths(self):
        return ['node_modules']

    def get_import_paths(self):
        return self.get_default_import_paths() + self.extra_import_paths

    def _get_import_paths_as_strlist(self):
        import_paths = []
        for path in self.get_import_paths():
            if isinstance(path, AbstractDjangoAppPath):
                path = path.abspath
            import_paths.append(path)
        return import_paths

    def install(self):
        """
        Installs the ``browserify`` NPM package.

        The package is installed with no version specified, so you
        probably want to freeze the version using the
        :class:`ievv_opensource.utils.ievvbuildstatic.npminstall.Plugin` plugin.
        """
        self.app.get_installer('npm').queue_install(
            'browserify', installtype='dev')
        if self.app.apps.is_in_production_mode():
            self.app.get_installer('npm').queue_install(
                'uglifyify', installtype='dev')

    def get_browserify_executable(self):
        return self.app.get_installer('npm').find_executable('browserify')

    def get_browserify_extra_args(self):
        """
        Get extra browserify args.

        Override this in subclasses to add transforms and such.
        """
        return []

    def get_browserify_production_args(self):
        return [
            '-t', '[', 'uglifyify', ']',
            '-g', 'uglifyify',
        ]

    def make_browserify_args(self):
        """
        Make browserify args list.

        Adds the following in the listed order:

        - ``-d`` if the ``sourcemap`` kwarg is ``True``.
        - the source file.
        - ``-d <destination file path>``.
        - whatever :meth:`.get_browserify_extra_args` returns.

        Should normally not be extended. Extend :meth:`.get_browserify_extra_args`
        instead.
        """
        args = []
        if self.sourcemap:
            args.append('-d')
        args.extend([
           self.get_sourcefile_path(),
           '-o', self.get_destinationfile_path(),
        ])
        args.extend(self.get_browserify_extra_args())
        if self.app.apps.is_in_production_mode():
            args.extend(self.get_browserify_production_args())
        return args

    def make_browserify_executable_environment(self):
        environment = os.environ.copy()
        environment.update({
            'NODE_PATH': ':'.join(self._get_import_paths_as_strlist())
        })
        return environment

    def post_run(self):
        """
        Called at the start of run(), after the initial command start logmessage,
        but before any other code is executed.

        Does nothing by default, but subclasses can hook in
        configuration code here if they need to generate config
        files, perform validation of file structure, etc.
        """

    def run(self):
        self.get_logger().command_start(
            'Running browserify with {sourcefile} as input and {destinationfile} as output'.format(
                sourcefile=self.get_sourcefile_path(),
                destinationfile=self.get_destinationfile_path()))
        self.post_run()
        executable = self.get_browserify_executable()
        destination_folder = self.app.get_destination_path(self.destinationfolder)
        if not os.path.exists(destination_folder):
            os.makedirs(destination_folder)
        try:
            self.run_shell_command(executable, args=self.make_browserify_args(),
                                   _cwd=self.app.get_source_path(),
                                   _env=self.make_browserify_executable_environment())
        except ShellCommandError:
            self.get_logger().command_error('browserify build FAILED!')
        else:
            self.get_logger().command_success('browserify build succeeded :)')

    def get_extra_watchfolder_paths(self):
        return map(self.app.get_source_path, self.extra_watchfolders)

    def get_watch_folders(self):
        """
        We only watch the folder where the javascript is located,
        so this returns the absolute path of the ``sourcefolder``.
        """
        folders = [self.app.get_source_path(self.sourcefolder)]
        if self.extra_watchfolders:
            folders.extend(self.get_extra_watchfolder_paths())
        return folders

    def get_watch_extensions(self):
        """
        Returns a list of the extensions to watch for in watch mode.

        Defaults to ``['js']``.

        Unless you have complex needs, it is probably easier to override
        this than overriding
        :meth:`~ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin.get_watch_regexes`.
        """
        return ['js']

    def get_watch_regexes(self):
        return ['^.+[.]({extensions})'.format(
            extensions='|'.join(self.get_watch_extensions()))]

    def __str__(self):
        return '{}({})'.format(super(Plugin, self).__str__(), self.sourcefolder)



from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _


class Tag(models.Model):
    """
    A single tag.

    A tag has a unique name, and data models is added to a tag via :class:`.TaggedObject`.
    """
    class Meta:
        verbose_name = _('Tag')
        verbose_name_plural = _('Tags')

    #: The label for the tag.
    taglabel = models.CharField(
        verbose_name=_('Tag'),
        max_length=30,
        unique=True,
        help_text=_('Maximum 30 characters.')
    )

    #: The tagtype is a way for applications to group tags by type.
    #: No logic is assigned to this field by default, other than
    #: that is is ``db_indexed``. The valid choices for the field
    #: is configured via the :setting:`IEVV_TAGFRAMEWORK_TAGTYPE_CHOICES`
    #: setting.
    tagtype = models.CharField(
        max_length=255,
        db_index=True,
        blank=True, null=False,
        default=''
    )

    @staticmethod
    def get_tagtype_choices():
        """
        Returns choices for :obj:`.tagtype`.

        This is not set as choices on the field (because changing that would
        trigger a migration), but it should be used in any form displaying
        tagtype.

        You configure the return value via the
        ``IEVV_TAGFRAMEWORK_TAGTYPE_CHOICES`` Django setting.
        """
        choices = getattr(settings, 'IEVV_TAGFRAMEWORK_SITE_TYPE_CHOICES',
                          [('', '')])
        for choice in choices:
            yield choice

    @classmethod
    def get_tagtype_valid_values(cls):
        """
        Returns an iterator over the choices returned by
        :meth:`.get_tagtype_choices`, but only the values (not the labels)
        as a flat list.
        """
        for value, label in cls.get_tagtype_choices():
            yield value

    def clean(self):
        tagtype_valid_values = set(Tag.get_tagtype_valid_values())
        if self.tagtype not in tagtype_valid_values:
            raise ValidationError({
                'tagtype': _('Must be one of: %(choices)s. Current value: %(value)s.') % {
                    'choices': ', '.join(repr(value) for value in tagtype_valid_values),
                    'value': repr(self.tagtype),
                }
            })


class TaggedObject(models.Model):
    """
    Represents a many-to-many relationship between any data model object and
    a :class:`.Tag`.
    """
    class Meta:
        verbose_name = _('Tagged object')
        verbose_name_plural = _('Tagged objects')

    #: The :class:`.Tag`.
    tag = models.ForeignKey(Tag, on_delete=models.CASCADE)

    #: The ContentType of the tagged object.
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)

    #: The ID of the tagged object.
    object_id = models.PositiveIntegerField()

    #: The GenericForeignKey using :obj:`~TaggedObject.content_type` and
    #: :obj:`~TaggedObject.object_id` to create a generic foreign key
    #: to the tagged object.
    content_object = GenericForeignKey('content_type', 'object_id')



import json
import logging
import os
import shutil
import time
from collections import OrderedDict

from django.conf import settings
from django.apps import apps
from ievv_opensource.utils.ievvbuildstatic import filepath
from ievv_opensource.utils.ievvbuildstatic.installers.yarn import YarnInstaller
from ievv_opensource.utils.ievvbuildstatic.watcher import WatchConfigPool
from ievv_opensource.utils.logmixin import LogMixin, Logger

from . import docbuilders


class App(LogMixin):
    """
    Configures how ``ievv buildstatic`` should build the static files for a Django app.
    """
    def __init__(self, appname, version, plugins,
                 sourcefolder='staticsources',
                 destinationfolder='static',
                 keep_temporary_files=False,
                 installers_config=None,
                 docbuilder_classes=None,
                 default_skipgroups=None,
                 default_includegroups=None,
                 appspecific_fields=None):
        """
        Parameters:
            appname: Django app label (I.E.: ``myproject.myapp``).
            plugins: Zero or more :class:`ievv_opensource.utils.ievvbuild.pluginbase.Plugin`
                objects.
            sourcefolder: The folder relative to the app root folder where
                static sources (I.E.: less, coffescript, ... sources) are located.
                Defaults to ``staticsources``.
            appspecific_fields: If your build-scenario requires more variables, add them in a dictionary here, and they
                will be included in the config.
                Defaults to ``{}``.
        """
        self.apps = None
        self.version = version
        self.appname = appname
        self.sourcefolder = sourcefolder
        self.destinationfolder = destinationfolder
        self.installers = {}
        self.plugins = []
        self.docbuilder_classes = docbuilder_classes or self.get_default_docbuilder_classes()
        self.keep_temporary_files = keep_temporary_files
        self.default_skipgroups = default_skipgroups or []
        self.default_includegroups = default_includegroups or []
        self.appspecific_fields = appspecific_fields or {}
        self.installers_config = self._make_installers_config(
            installers_config_overrides=installers_config)
        for plugin in plugins:
            self.add_plugin(plugin)

    def _make_installers_config(self, installers_config_overrides):
        installers_config = {
            'npm': {
                'installer_class': YarnInstaller
            },
        }
        installers_config_overrides = installers_config_overrides or {}
        for alias, config in installers_config_overrides.items():
            overrides = installers_config_overrides.get(alias, {})
            if alias not in installers_config:
                installers_config[alias] = {}
            installers_config[alias].update(overrides)
        return installers_config

    def get_options_classes(self):
        options_classes = []
        for installer_config in self.installers_config.values():
            options_classes.append(installer_config['installer_class'])
        for plugin in self.plugins:
            options_classes.append(plugin.__class__)
        return options_classes

    def add_plugin(self, plugin):
        """
        Add a :class:`ievv_opensource.utils.ievvbuildstatic.lessbuild.Plugin`.
        """
        plugin.app = self
        self.plugins.append(plugin)

    def iterplugins(self, skipgroups=None, includegroups=None):
        skipgroups = skipgroups or self.default_skipgroups
        includegroups = includegroups or self.default_includegroups
        skipgroups = {group for group in skipgroups if group not in includegroups}
        for plugin in self.plugins:
            if includegroups and plugin.group not in includegroups:
                continue
            if plugin.group and plugin.group in skipgroups:
                continue
            yield plugin

    def run(self, skipgroups=None, includegroups=None):
        """
        Run :meth:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin.run`
        for all plugins within the app.
        """
        for plugin in self.iterplugins(skipgroups=skipgroups, includegroups=includegroups):
            plugin.runwrapper()

    def _make_json_appconfig_dict(self):
        return {
            'appname': self.appname,
            'version': self.version,
            'sourcefolder': self.get_source_path(),
            'destinationfolder': self.get_destination_path(),
            'keep_temporary_files': self.keep_temporary_files,
            'is_in_production_mode': self.apps.is_in_production_mode(),
            'static_url': self.get_static_url(),
            'appspecific_fields': self.appspecific_fields
        }

    def _get_json_appconfig_path(self):
        return self.get_source_path('ievv_buildstatic.appconfig.json')

    def _save_json_appconfig(self, appconfig_dict):
        config_path = self._get_json_appconfig_path()
        self.get_logger().debug('Creating {config_path}'.format(config_path=config_path))
        open(config_path, 'w').write(
            json.dumps(appconfig_dict, indent=2, sort_keys=True)
        )

    def add_pluginconfig_to_json_config(self, plugin_name, config_dict):
        appconfig_dict = json.loads(open(self._get_json_appconfig_path(), 'r').read())
        appconfig_dict[plugin_name] = config_dict
        self._save_json_appconfig(appconfig_dict=appconfig_dict)

    def install(self, skipgroups=None, includegroups=None):
        """
        Run :meth:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin.install`
        for all plugins within the app.
        """
        self._save_json_appconfig(appconfig_dict=self._make_json_appconfig_dict())
        for alias in self.installers_config.keys():
            installer = self.get_installer(alias=alias)
            installer.initialize()
        for plugin in self.iterplugins(skipgroups=skipgroups, includegroups=includegroups):
            plugin.install()
        for installer in self.installers.values():
            installer.install()
        for plugin in self.iterplugins(skipgroups=skipgroups, includegroups=includegroups):
            plugin.post_install()

    def get_app_config(self):
        """
        Get the AppConfig for the Django app.
        """
        if not hasattr(self, '_app_config'):
            self._app_config = apps.get_app_config(self.appname)
        return self._app_config

    def get_appfolder(self):
        """
        Get the absolute path to the Django app root folder.
        """
        return self.get_app_config().path

    def get_app_path(self, apprelative_path):
        """
        Returns the path to the directory joined with the
        given ``apprelative_path``.
        """
        return os.path.join(self.get_appfolder(), apprelative_path)

    def _relative_or_absolute_path_to_absolute_path(self, relative_path_root, pathlist):
        if len(pathlist) == 1 and isinstance(pathlist[0], filepath.FilePathInterface):
            return pathlist[0].abspath
        else:
            if pathlist:
                return os.path.join(relative_path_root, *pathlist)
            else:
                return relative_path_root

    def get_static_url(self):
        """
        Returns the static-url for the output-directory.
        This url is needed for webpack when using code-splitting during javascript-build, as it will be used by the
        frontend to load the chunks when/if needed.

        This will typically be something like:

            `/static/myapp/42/`

        where `/static/` is `settings.STATIC_URL`, `myapp` is the appname, and `42` is the version.
        """
        return f'{settings.STATIC_URL}{self.appname}/{self.version}/'

    def get_source_path(self, *path):
        """
        Returns the absolute path to a folder within the source
        folder of this app or another app.

        Examples:

            Get the source path for a coffeescript file::

                self.get_source_path('mylib', 'app.coffee')

            Getting the path of a source file within another app using
            a :class:`ievv_opensource.utils.ievvbuildstatic.filepath.SourcePath`
            object (a subclass of :class:`ievv_opensource.utils.ievvbuildstatic.filepath.FilePathInterface`)
            as the path::

                self.get_source_path(
                    ievvbuildstatic.filepath.SourcePath('myotherapp', 'scripts', 'typescript', 'app.ts'))


        Args:
            *path: Zero or more strings to specify a path relative to the source folder of this app -
                same format as :func:`os.path.join`.
                A single :class:`ievv_opensource.utils.ievvbuildstatic.filepath.FilePathInterface`
                object to specify an absolute path.
        """
        sourcefolder = os.path.join(self.get_app_path(self.sourcefolder), self.appname)
        return self._relative_or_absolute_path_to_absolute_path(relative_path_root=sourcefolder,
                                                                pathlist=path)

    def make_source_relative_path(self, *path):
        return os.path.relpath(self.get_source_path(*path), start=self.get_source_path())

    def get_destination_path(self, *path, **kwargs):
        """
        Returns the absolute path to a folder within the destination
        folder of this app or another app.

        Examples:

            Get the destination path for a coffeescript file - extension
            is changed from ``.coffee`` to ``.js``::

                self.get_destination_path('mylib', 'app.coffee', new_extension='.js')

            Getting the path of a destination file within another app using
            a :class:`ievv_opensource.utils.ievvbuildstatic.filepath.SourcePath`
            object (a subclass of :class:`ievv_opensource.utils.ievvbuildstatic.filepath.FilePathInterface`)
            as the path::

                self.get_destination_path(
                    ievvbuildstatic.filepath.DestinationPath(
                        'myotherapp', '1.1.0', 'scripts', 'typescript', 'app.ts'),
                    new_extension='.js')

        Args:
            path: Path relative to the source folder.
                Same format as ``os.path.join()``.
                A single :class:`ievv_opensource.utils.ievvbuildstatic.filepath.FilePathInterface`
                object to specify an absolute path.
            new_extension: A new extension to give the destination path.
                See example below.

        """
        new_extension = kwargs.get('new_extension', None)
        destinationfolder = os.path.join(
            self.get_app_path(self.destinationfolder), self.appname, self.version)
        absolute_path = self._relative_or_absolute_path_to_absolute_path(relative_path_root=destinationfolder,
                                                                         pathlist=path)
        if new_extension is not None:
            path, extension = os.path.splitext(absolute_path)
            absolute_path = '{}{}'.format(path, new_extension)
        return absolute_path

    def watch(self, skipgroups=None, includegroups=None):
        """
        Start a watcher thread for each plugin.
        """
        watchconfigs = []
        for plugin in self.iterplugins(skipgroups=skipgroups, includegroups=includegroups):
            watchconfig = plugin.watch()
            if watchconfig:
                watchconfigs.append(watchconfig)
        return watchconfigs

    def iterinstallers(self):
        return self.installers.values()

    def get_installer(self, alias):
        """
        Get an instance of the installer configured with the provided ``alias``.

        Parameters:
            alias: A subclass of
                .
        Returns:
            ievv_opensource.utils.ievvbuildstatic.installers.base.AbstractInstaller: An instance
                of the requested installer.
        """
        installer_class = self.installers_config[alias]['installer_class']
        if installer_class.name not in self.installers:
            installer = installer_class(app=self)
            self.installers[installer_class.name] = installer
        return self.installers[installer_class.name]

    def get_logger_name(self):
        return '{}.{}'.format(self.apps.get_logger_name(), self.appname)

    def get_loglevel(self):
        return self.apps.loglevel

    def get_temporary_build_directory_path(self, *path):
        return self.get_source_path('ievvbuildstatic_temporary_build_directory', *path)

    def make_temporary_build_directory(self, name):
        """
        Make a temporary directory that you can use for building something.

        Returns:
            str: The absolute path of the new directory.
        """
        self._delete_temporary_build_directory(name)
        temporary_directory_path = self.get_temporary_build_directory_path(name)
        os.makedirs(temporary_directory_path)
        return temporary_directory_path

    def _delete_temporary_build_directory(self, name):
        """
        Delete a temporary directory created with :meth:`.make_temporary_build_directory`.
        """
        temporary_directory_path = self.get_temporary_build_directory_path(name)
        if os.path.exists(temporary_directory_path):
            shutil.rmtree(temporary_directory_path)
        base_temporary_directory = self.get_temporary_build_directory_path()
        if os.path.exists(base_temporary_directory) and len(os.listdir(base_temporary_directory)) == 0:
            shutil.rmtree(base_temporary_directory)

    def get_default_docbuilder_classes(self):
        return [
            docbuilders.npm_docbuilder.NpmDocBuilder
        ]

    def get_docbuilder(self):
        for docbuilder_class in self.docbuilder_classes:
            docbuilder = docbuilder_class(app=self)
            if docbuilder.is_available():
                return docbuilder
        return None

    def build_docs(self, output_directory):
        docbuilder = self.get_docbuilder()
        if docbuilder:
            docbuilder.build_docs(output_directory=os.path.join(output_directory, self.appname))
        else:
            self.get_logger().debug(
                'No docbuilders found for {appname}. Tried the following docbuilders: {docbuilders}'.format(
                    appname=self.appname,
                    docbuilders=', '.join(
                        docbuilder_class.name
                        for docbuilder_class in self.docbuilder_classes)
                ))

    def add_deferred_success(self, message):
        self.apps.add_deferred_success('[{}] {}'.format(self.appname, message))

    def add_deferred_warning(self, message):
        self.apps.add_deferred_warning('[{}] {}'.format(self.appname, message))


class Apps(LogMixin):
    """
    Basically a list around :class:`.App` objects.
    """
    MODE_DEVELOP = 'develop'
    MODE_PRODUCTION = 'production'

    def __init__(self, *apps, **kwargs):
        """
        Parameters:
            apps: :class:`.App` objects to add initially. Uses :meth:`.add_app` to add the apps.
        """
        self.apps = OrderedDict()
        self.loglevel = Logger.DEBUG
        self.command_error_message = None
        self.help_header = kwargs.pop('help_header', None)
        self.mode = self.MODE_DEVELOP
        self._warnings = []
        self._successes = []
        self.options = {}
        for app in apps:
            self.add_app(app)

    def get_loglevel(self):
        return self.loglevel

    def get_command_error_message(self):
        return self.command_error_message

    def add_app(self, app):
        """
        Add an :class:`.App`.
        """
        app.apps = self
        self.apps[app.appname] = app

    def get_app(self, appname):
        """
        Get app by appname.
        """
        return self.apps[appname]

    def install(self, appnames=None, skipgroups=None, includegroups=None):
        """
        Run :meth:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin.install`
        for all plugins within all :class:`apps <.App>`.
        """
        for app in self.iterapps(appnames=appnames):
            app.install(skipgroups=skipgroups, includegroups=includegroups)

    def log_help_header(self):
        if self.help_header:
            self.get_logger().infobox(self.help_header)

    def validate_appnames(self, appnames):
        for appname in appnames:
            if appname not in self.apps:
                raise ValueError('Invalid appname: {}'.format(appname))

    def iterapps(self, appnames=None):
        """
        Get an interator over the apps.
        """
        for app in self.apps.values():
            include = not appnames or app.appname in appnames
            if include:
                yield app

    def run(self, appnames=None, skipgroups=None, includegroups=None):
        """
        Run :meth:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin.run`
        for all plugins within all :class:`apps <.App>`.
        """
        for app in self.iterapps(appnames=appnames):
            app.run(skipgroups=skipgroups, includegroups=includegroups)
        self.log_deferred_messages()

    def watch(self, appnames=None, skipgroups=None, includegroups=None):
        """
        Start watcher threads for all folders that at least one
        :class:`plugin <ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin>`
        within any of the :class:`apps <.App>` has configured to be watched for changes.

        Blocks until ``CTRL-c`` is pressed.
        """
        watchconfigpool = WatchConfigPool()
        for app in self.iterapps(appnames=appnames):
            watchconfigpool.extend(app.watch(skipgroups=skipgroups, includegroups=includegroups))
        all_observers = watchconfigpool.watch()
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            for observer in all_observers:
                observer.stop()

        for observer in all_observers:
            observer.join()

    def build_docs(self, output_directory, appnames=None):
        for app in self.iterapps(appnames=appnames):
            app.build_docs(output_directory=output_directory)

    def get_logger_name(self):
        return 'ievvbuildstatic'

    def __configure_shlogger(self, loglevel, handler):
        shlogger = logging.getLogger('sh.command')
        shlogger.setLevel(loglevel)
        shlogger.addHandler(handler)
        shlogger.propagate = False

    def configure_logging(self, loglevel=None, shlibrary_loglevel=logging.WARNING,
                          command_error_message=None):
        handler = logging.StreamHandler()
        self.__configure_shlogger(loglevel=shlibrary_loglevel,
                                  handler=handler)
        self.loglevel = loglevel
        self.command_error_message = command_error_message

    def set_development_mode(self):
        self.mode = self.MODE_DEVELOP

    def set_production_mode(self):
        self.mode = self.MODE_PRODUCTION

    def is_in_production_mode(self):
        return self.mode == self.MODE_PRODUCTION

    def add_deferred_warning(self, message):
        self._warnings.append(message)

    def add_deferred_success(self, message):
        self._successes.append(message)

    def log_deferred_messages(self):
        if self._successes:
            self.get_logger().success('Success messages:')
            for message in set(self._successes):
                self.get_logger().success('- {}'.format(message))
        if self._warnings:
            self.get_logger().warning('Warnings:')
            for message in set(self._warnings):
                self.get_logger().warning('- {}'.format(message))
        self._successes = []
        self._warnings = []

    def add_cli_arguments(self, parser):
        options_classes = set()
        for app in self.iterapps():
            options_classes.update(app.get_options_classes())
        for option_class in options_classes:
            option_class.add_cli_arguments(parser=parser)

    def set_options(self, options):
        self.options = options



from django.conf import settings
from django.db import models


def collate(field_name, collation=None):
    """
    Enforces collation specified by the given paramater, or the
    setting `IEVV_DB_POSTGRES_COLLATION`. Wrapping `django.db.models.Func`-class
    and generates a function expression with collation.

    Note::
        Returns the field name if collation is not defined in settings,
        which results in the default Django order_by behaviour.

    Args:
        field_name (str): The field name.
        collation (str): The collation to use. 'E.g en_US.UTF-8'
    """

    # Get collation from settings.
    if not collation:
        collation = getattr(settings, 'IEVV_DB_POSTGRES_COLLATION', None)

    # If collation is not defined in settings, return the field name.
    # This will result in the default wayt to do order_by (...order_by('<field_name>'))
    if not collation:
        return field_name

    # Determine if descending order should be used.
    # Setting desc_order to True if thats the case.
    desc_order = False
    if field_name.startswith('-'):
        field_name = field_name.split('-')[1]
        desc_order = True

    collate_function = models.Func(
        field_name,
        function=collation,
        template='(%(expressions)s) COLLATE "%(function)s"'
    )

    if desc_order:
        return collate_function.desc()
    return collate_function.asc()



from django import test

from ievv_opensource.utils import choices_with_meta


class TestChoice(test.TestCase):
    def test_init_label_fallback_to_value(self):
        choice = choices_with_meta.Choice(value='test')
        self.assertEqual('test', choice.label)

    def test_get_short_label(self):
        choice = choices_with_meta.Choice(value='test', label='Test')
        self.assertEqual('Test', choice.get_short_label())

    def test_get_long_label_without_description(self):
        choice = choices_with_meta.Choice(value='test', label='Test')
        self.assertEqual('Test', choice.get_long_label())

    def test_get_long_label_with_description(self):
        choice = choices_with_meta.Choice(value='test', label='Test', description='A description')
        self.assertEqual('Test - A description', choice.get_long_label())

    def test_value_to_attributename(self):
        self.assertEqual('TEST',
                         choices_with_meta.Choice(value='test').attributename)
        self.assertEqual('MULTI_WORD',
                         choices_with_meta.Choice(value='multi word').attributename)
        self.assertEqual('LINE_SEPARATED_WORDS',
                         choices_with_meta.Choice(value='line-separated-words').attributename)
        self.assertEqual('_0',
                         choices_with_meta.Choice(value=0).attributename)

    def test_custom_attributename(self):
        self.assertEqual('first',
                         choices_with_meta.Choice(value=0, attributename='first').attributename)


class TestChoicesWithMeta(test.TestCase):
    def test_add_adds_to_choices(self):
        choices = choices_with_meta.ChoicesWithMeta()
        choice = choices_with_meta.Choice(value='test1')
        choices.add(choice)
        self.assertEqual(choice, choices.choices['test1'])

    def test_add_adds_to_attributes(self):
        choices = choices_with_meta.ChoicesWithMeta()
        choice = choices_with_meta.Choice(value='test1')
        choices.add(choice)
        self.assertEqual(choice, choices.TEST1)

    def test_getitem_value_exists(self):
        choices = choices_with_meta.ChoicesWithMeta()
        choice = choices_with_meta.Choice(value='test1')
        choices.add(choice)
        self.assertEqual(choice, choices['test1'])

    def test_getitem_value_does_not_exist(self):
        choices = choices_with_meta.ChoicesWithMeta()
        with self.assertRaises(KeyError):
            choices['somevalue']

    def test_get_by_value_value_exists(self):
        choices = choices_with_meta.ChoicesWithMeta()
        choice = choices_with_meta.Choice(value='test1')
        choices.add(choice)
        self.assertEqual(choice, choices.get_by_value('test1'))

    def test_get_by_value_value_does_not_exist(self):
        choices = choices_with_meta.ChoicesWithMeta()
        self.assertIsNone(choices.get_by_value('somevalue'))

    def test_get_by_value_value_does_not_exist_custom_fallback(self):
        choices = choices_with_meta.ChoicesWithMeta()
        self.assertEqual('fall back', choices.get_by_value('somevalue', 'fall back'))

    def test_getattr_invalid_value(self):
        choices = choices_with_meta.ChoicesWithMeta()
        with self.assertRaises(AttributeError):
            choices.INVALID

    def test_getattr_valid_value(self):
        choices = choices_with_meta.ChoicesWithMeta()
        choice = choices_with_meta.Choice(value='test1')
        choices.add(choice)
        self.assertEqual(choice, choices.TEST1)

    def test_contains(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1'))
        self.assertTrue('test1' in choices)
        self.assertFalse('test2' in choices)

    def test_len(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1'),
            choices_with_meta.Choice(value='test2'))
        self.assertEqual(2, len(choices))

    def test_get_choice_at_index(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1'),
            choices_with_meta.Choice(value='test2'))
        self.assertEqual('test1', choices.get_choice_at_index(0).value)
        self.assertEqual('test2', choices.get_choice_at_index(1).value)

    def test_get_choice_at_index_invalid_index(self):
        choices = choices_with_meta.ChoicesWithMeta()
        with self.assertRaises(IndexError):
            choices.get_choice_at_index(0)

    def test_get_first_choice(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1'),
            choices_with_meta.Choice(value='test2'))
        self.assertEqual('test1', choices.get_first_choice().value)

    def test_get_first_choice_none(self):
        choices = choices_with_meta.ChoicesWithMeta()
        self.assertIsNone(choices.get_first_choice())

    def test_init(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1', label='Test1'),
            choices_with_meta.Choice(value='test2', label='Test2'))
        self.assertEqual('Test1', choices['test1'].label)
        self.assertEqual('Test2', choices['test2'].label)

    def test_itervalues(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1'),
            choices_with_meta.Choice(value='test2'))
        self.assertEqual(['test1', 'test2'], list(choices.itervalues()))

    def test_get_values_as_list(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1'),
            choices_with_meta.Choice(value='test2'))
        self.assertEqual(['test1', 'test2'], choices.get_values_as_list())

    def test_iterchoices(self):
        choice1 = choices_with_meta.Choice(value='test1')
        choice2 = choices_with_meta.Choice(value='test2')
        choices = choices_with_meta.ChoicesWithMeta(choice1, choice2)
        self.assertEqual([choice1, choice2], list(choices.iterchoices()))

    def test_iter_as_django_choices_short(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1', label='Test1'),
            choices_with_meta.Choice(value='test2', label='Test2'))
        self.assertEqual(
            [
                ('test1', 'Test1'),
                ('test2', 'Test2'),
            ],
            list(choices.iter_as_django_choices_short()))

    def test_iter_as_django_choices_long(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1', label='Test1', description='A description'),
            choices_with_meta.Choice(value='test2', label='Test2'))
        self.assertEqual(
            [
                ('test1', 'Test1 - A description'),
                ('test2', 'Test2'),
            ],
            list(choices.iter_as_django_choices_long()))

    def test_get_values_as_commaseparated_string(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1'),
            choices_with_meta.Choice(value='test2'))
        self.assertEqual('test1, test2', choices.get_values_as_commaseparated_string())

    def test_str(self):
        choices = choices_with_meta.ChoicesWithMeta(
            choices_with_meta.Choice(value='test1'),
            choices_with_meta.Choice(value='test2'))
        self.assertEqual('ChoicesWithMeta(test1, test2)', str(choices))

    def test_get_default_choices_added_in_init(self):
        class MyChoicesWithMeta(choices_with_meta.ChoicesWithMeta):
            def get_default_choices(self):
                return [
                    choices_with_meta.Choice(value='default1'),
                    choices_with_meta.Choice(value='default2')
                ]

        choices = MyChoicesWithMeta(
            choices_with_meta.Choice(value='test1'))
        self.assertEqual(
            ['default1', 'default2', 'test1'],
            choices.get_values_as_list()
        )





class LazyStatic(object):
    """
    Works just like the ``static`` template tag, but this is lazy
    (only evaluates in __str__), and can safely be used in settings.py etc.
    """
    def __init__(self, path):
        self.path = path

    def __str__(self):
        from django.templatetags.static import static
        return static(self.path)

    def __repr__(self):
        return 'LazyStatic({})'.format(self.path)



from django.http import HttpResponseRedirect
from django.views import View

from ievv_opensource.ievv_i18n_url import i18n_url_utils


class RedirectToLanguagecodeView(View):
    def dispatch(self, request, languagecode, path):
        handler = i18n_url_utils.get_handler()
        return HttpResponseRedirect(handler.build_absolute_url(f'/{path}', languagecode=languagecode))



import sys

from ievv_opensource.utils.ievvdevrun.runnables import base


class RunnableThread(base.ShellCommandRunnableThread):
    """
    Celery worker runnable thread.

    Examples:

        You can just add it to your Django development settings with::

            IEVVTASKS_DEVRUN_RUNNABLES = {
                'default': ievvdevrun.config.RunnableThreadList(
                    ievvdevrun.runnables.celery_worker.RunnableThread(
                        app='myproject')
                )
            }

        And you can make it not restart on crash with::

            IEVVTASKS_DEVRUN_RUNNABLES = {
                'default': ievvdevrun.config.RunnableThreadList(
                    ievvdevrun.runnables.celery_worker.RunnableThread(
                        app='myproject',
                        autorestart_on_crash=False)
                )
            }

    """
    default_autorestart_on_crash = True

    def __init__(self, app, loglevel='debug'):
        """
        Args:
            app: The Celery app to start - the python path to the module where you
                import your celery app in ``__init__.py``.
            loglevel: The Celery loglevel (``-l`` command line argument). Defaults to ``"debug"``.
        """
        self.app = app
        self.loglevel = loglevel
        super(RunnableThread, self).__init__()

    def get_logger_name(self):
        return 'Celery worker'

    def get_command_config(self):
        return {
            'executable': 'celery',
            'args': ['-A', self.app, 'worker', '-l', self.loglevel]
        }



import os

from ievv_opensource.utils.logmixin import LogMixin
from ievv_opensource.utils.shellcommandmixin import ShellCommandMixin


class AbstractDocBuilder(LogMixin, ShellCommandMixin):
    """
    Base class for documentation builders.

    Each installer defines most of their own API, the only thing they
    have in common is a reference to their
    :class:`ievv_opensource.utils.ievvbuildstatic.config.App`,
    a :obj:`~.AbstractDocBuilder.name` and the methods
    defined in :class:`ievv_opensource.utils.logmixin.LogMixin`
    and :class:`ievv_opensource.utils.ievvbuildstatic.shellcommand.ShellCommandMixin`.

    An app runs all the doc builders in order. The first doc builder
    where :meth:`~.AbstractDocBuilder.is_available` returns ``True``
    is used for the app.
    """

    #: The name of the docbuilder.
    name = None

    def __init__(self, app):
        """
        Parameters:
            app: The :class:`ievv_opensource.utils.ievvbuildstatic.config.App` where
                this docbuilder object was instantiated.
        """
        self.app = app

    def get_logger_name(self):
        return '{}.{}'.format(self.app.get_logger_name(), self.name)

    def get_loglevel(self):
        return self.app.apps.loglevel

    def is_available(self):
        """
        Is this docbuilder available in the app?
        """
        return False

    def build_docs(self, output_directory):
        raise NotImplementedError()



import os
from ievv_opensource.utils.desktopnotifications.base import AbstractNotification

try:
    from shlex import quote
except ImportError:
    from pipes import quote


class Notification(AbstractNotification):
    def show_message(self, title, message):
        os.system('notify-send {} {}'.format(quote(title), quote(message)))



class OptionsMixin(object):
    def get_option(self, key, fallback=None):
        return self.app.apps.options.get(key, fallback)

    def is_in_production_mode(self):
        return self.app.apps.is_in_production_mode()

    @classmethod
    def add_cli_arguments(cls, parser):
        """
        Add CLI arguments.

        The arguments are forwarder to __init__ via the


        Args:
            parser: argparse parser.
        """



from collections import OrderedDict

from ievv_opensource.utils.shellcommandmixin import ShellCommandError
from .abstract_npm_installer import AbstractNpmInstaller


class NpmInstallerError(Exception):
    pass


class PackageJsonDoesNotExist(NpmInstallerError):
    pass


class NpmInstaller(AbstractNpmInstaller):
    """
    NPM installer.
    """
    name = 'npminstall'
    optionprefix = 'npm'

    def _run_npm(self, args):
        self.run_shell_command('npm',
                               args=args,
                               _cwd=self.app.get_source_path())

    def __init__(self, *args, **kwargs):
        super(NpmInstaller, self).__init__(*args, **kwargs)
        self.queued_packages = OrderedDict()

    def log_shell_command_stderr(self, line):
        if 'npm WARN package.json' in line:
            return
        super(NpmInstaller, self).log_shell_command_stderr(line)

    def install_packages_from_packagejson(self):
        try:
            self.run_shell_command('npm',
                                   args=['install'],
                                   _cwd=self.app.get_source_path())
        except ShellCommandError:
            self.get_logger().command_error('npm install FAILED!')
            raise SystemExit()

    def install_npm_package(self, package, properties):
        package_spec = package
        if properties['version']:
            package_spec = '{package}@{version}'.format(
                package=package, version=properties['version'])
        args = ['install', package_spec]
        if properties['installtype'] is None:
            args.append('--save')
        else:
            args.append('--save-{}'.format(properties['installtype']))
        try:
            self.run_shell_command('npm',
                                   args=args,
                                   _cwd=self.app.get_source_path())
        except ShellCommandError:
            self.get_logger().command_error(
                'npm install {package} (properties: {properties!r}) FAILED!'.format(
                    package=package, properties=properties))
            raise SystemExit()

    def run_packagejson_script(self, script, args=None):
        args = args or []
        self.run_shell_command('npm',
                               args=['run', script] + args,
                               _cwd=self.app.get_source_path())

    def link_package(self, packagename):
        self._run_npm(args=['link', packagename])

    def unlink_package(self, packagename):
        self._run_npm(args=['unlink', packagename])

    def install(self):
        self.get_logger().command_start(
            'Installing npm packages for {}'.format(self.app.get_source_path()))
        if self.get_option('clean_node_modules', False):
            self.remove_node_modules_directory()
        if not self.packagejson_exists():
            self.create_packagejson()
        self.install_packages_from_packagejson()
        self.install_queued_packages()
        self.handle_linked_packages()
        self.get_logger().command_success('Install npm packages succeeded :)')
        self.add_deferred_success('Install npm packages succeeded :)')



from django.conf import settings


def get_fallback_base_url_setting():
    """Get the IEVV_I18N_URL_FALLBACK_BASE_URL.

    Raises:
        Exception: With a nice description if the setting is missing.

    Returns:
        str: The value of the setting
    """
    fallback_base_url = getattr(settings, 'IEVV_I18N_URL_FALLBACK_BASE_URL', None)
    if fallback_base_url:
        return fallback_base_url
    raise Exception(
        'The IEVV_I18N_URL_FALLBACK_BASE_URL setting is required. '
        'Refer to the docs for ievv_i18n_url (in ievv_opensource) for more info.')



import json
import os

from . import browserify_jsbuild


class Plugin(browserify_jsbuild.Plugin):
    """
    Browserify javascript babel build plugin.

    Examples:

        Simple example::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.browserify_babelbuild.Plugin(
                            sourcefile='app.js',
                            destinationfile='app.js',
                        ),
                    ]
                )
            )

        Custom source folder example::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.browserify_babelbuild.Plugin(
                            sourcefolder=os.path.join('scripts', 'javascript', 'api'),
                            sourcefile='api.js',
                            destinationfile='api.js',
                        ),
                    ]
                )
            )

    """
    name = 'browserify_babelbuild'

    def __init__(self, echmascript_version='es2015', autocreate_babelrc=True, **kwargs):
        """
        Parameters:
            echmascript_version: The echmascript version to use. Defaults to ``"es2015"``.
            autocreate_babelrc (bool):
            **kwargs: Kwargs for
                :class:`ievv_opensource.utils.ievvbuildstatic.browserify_jsbuild.Plugin`.
        """
        self.echmascript_version = echmascript_version
        self.autocreate_babelrc = autocreate_babelrc
        super(Plugin, self).__init__(**kwargs)

    def install(self):
        """
        Installs the ``babelify`` and ``babel-preset-<echmascript_version kwarg>``
        NPM packages in addition to the packages installed by
        :meth:`ievv_opensource.utils.ievvbuildstatic.browserify_jsbuild.Plugin.install`.

        The packages are installed with no version specified, so you
        probably want to freeze the versions using the
        :class:`ievv_opensource.utils.ievvbuildstatic.npminstall.Plugin` plugin.
        """
        super(Plugin, self).install()
        self.app.get_installer('npm').queue_install(
            'babelify')
        self.app.get_installer('npm').queue_install(
            'babel-preset-{}'.format(self.echmascript_version))

    def get_babel_presets(self):
        """
        Get a list of babelify presets.

        This is the presets that go into ``<HERE>`` in
        ``babelify -t [ babelify --presets [ <HERE> ] ]``.

        Defaults to ``["<the echmascript_version kwarg>"]``.
        """
        presets = [self.echmascript_version]
        return presets

    def get_browserify_extra_args(self):
        return ['-t', '[', 'babelify', ']']

    def get_babelrc_path(self):
        return self.app.get_source_path('.babelrc')

    def babelrc_exists(self):
        return os.path.exists(self.get_babelrc_path())

    def make_babelrc_dict(self):
        return {
            "presets": self.get_babel_presets()
        }

    def create_babelrc(self):
        open(self.get_babelrc_path(), 'w').write(
            json.dumps(
                self.make_babelrc_dict(),
                indent=2,
                sort_keys=True
            ))

    def post_run(self):
        if self.autocreate_babelrc:
            self.create_babelrc()
        elif not self.babelrc_exists():
            self.get_logger().command_error(
                '{babelrc_path} does not exist.'
                'When configured with autocreate_babelrc=False, we expect '
                'you to create the .babelrc.'.format(babelrc_path=self.get_babelrc_path()))
            raise SystemExit()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import math
import re

from django.conf import settings

from ievv_opensource.utils.singleton import Singleton
import gsm0338  # registers the gsm03.38 text encoder used in get_sms_text_length


class AbstractSmsBackend(object):
    """
    Base class for SMS backends.

    An instance of this class is created each time an SMS is created.

    This means that you can store temporary information related
    to building the SMS on ``self``.

    Example (simple print SMS backend)::

        class PrintBackend(sms_registry.AbstractSmsBackend):
            @classmethod
            def get_backend_id(cls):
                return 'print'

            def send(self):
                print(
                    'Phone number: {phone_number}. '
                    'Message: {message}'.format(
                        phone_number=self.cleaned_phone_number,
                        message=self.cleaned_message
                    )
                )

    To use the PrintBackend, add it to the registry via an AppConfig for your
    Django app::

        from django.apps import AppConfig

        class MyAppConfig(AppConfig):
            name = 'myapp'

            def ready(self):
                from ievv_opensource.ievv_sms import sms_registry
                from myapp import sms_backends
                sms_registry.Registry.get_instance().add(sms_backends.PrintBackend)

    Now you can use the backend to send an SMS::

        from ievv_opensource.ievv_sms import sms_registry
        sms_registry.Registry.get_instance().send(
            phone_number='12345678',
            message='This is a test',
            backend_id='print')

    You can also set the backend as the default backend
    for SMS sending by adding ``IEVV_SMS_DEFAULT_BACKEND_ID = 'print'`` to
    your django settings. With this setting you can call send()
    without the ``backend_id`` argument, and the SMS will be sent
    with the print backend.
    """
    STRIP_WHITESPACE_PATTERN = re.compile(r'\s+')

    @classmethod
    def get_max_length(cls):
        return 153 * 6

    @classmethod
    def get_sms_text_length(cls, text):
        return len(text.encode('gsm03.38', errors='replace'))

    @classmethod
    def get_part_count(cls, text):
        sms_text_length = cls.get_sms_text_length(text)
        if sms_text_length <= 160:
            return 1
        return int(math.ceil(sms_text_length / 153))

    @classmethod
    def get_backend_id(cls):
        """
        The ID this backend will get get in the :class:`.Registry` singleton.

        Defaults to the full python path for the class.
        """
        return '{}.{}'.format(cls.__module__, cls.__name__)

    @classmethod
    def validate_backend_setup(cls):
        """
        Validate backend setup (required settings, etc).

        Raises:
            .SmsBackendSetupError: When the setup validation fails.
        """

    def __init__(self, phone_number, message, send_as=None, **kwargs):
        """
        All the arguments are forwarded from :meth:`.Registry.send` /
        :meth:`.Registry.make_backend_instance`.

        Args:
            phone_number (str): The phone number to send the message to.
            message (str): The message to send.
            **kwargs: Extra kwargs. Both for future proofing, and
                to make it possible for backends to support extra
                kwargs.
        """
        self.phone_number = phone_number
        self.message = message
        self.send_as = send_as
        self.kwargs = kwargs

    def clean_phone_number(self, phone_number):
        """
        Clean/validate the phone number.

        By default this does nothing. It is here for backends
        that need to format phone numbers in a specific way.

        Args:
            phone_number (str): The phone number to clean.

        Raises:
             django.core.exceptions.ValidationError: If validation of the phone number fails.

        Returns:
            str: The cleaned phone number.
        """
        return phone_number

    def clean_message(self, message):
        """
        Clean/validate the message.

        By default this does nothing. It is here for backends
        that need to format or validate the message in a specific way.

        Args:
            message (str): The message to clean.

        Raises:
             django.core.exceptions.ValidationError: If validation of the message fails.

        Returns:
            str: The cleaned message.
        """
        return message

    def clean(self):
        """
        Clean the phone number, message, and kwargs.

        Calls :meth:`.clean_phone_number` and :meth:`.clean_message`.

        If you need to clean extra kwargs, you should override this method,
        but make sure you call ``super().clean()``.

        Raises:
             django.core.exceptions.ValidationError: If validation fails.
        """
        self.cleaned_phone_number = self.clean_phone_number(phone_number=self.phone_number)
        self.cleaned_message = self.clean_message(message=self.message)

    def send(self):
        """
        Send the message.

        Must be overridden in subclasses.

        Should send ``self.cleaned_message`` to ``self.cleaned_phone_number``.
        """
        raise NotImplementedError()

    def __str__(self):
        return '{}.{}(phone_number={!r}, message={!r}, kwargs={!r})'.format(
            self.__class__.__module__, self.__class__.__name__,
            self.phone_number, self.message, self.kwargs
        )

    def __repr__(self):
        return str(self)


class Registry(Singleton):
    """
    Registry of :class:`.AbstractSmsBackend` objects.
    """

    def __init__(self):
        super(Registry, self).__init__()
        self._backend_class_map = {}

    def add(self, backend_class):
        """
        Add the given ``backend_class`` to the registry.

        Parameters:
            backend_class: A subclass of :class:`.AbstractSmsBackend`.
        """
        backend_id = backend_class.get_backend_id()
        if backend_id in self._backend_class_map:
            raise ValueError('{!r} is already in the SMS backend registry.'.format(
                backend_id))
        self._backend_class_map[backend_id] = backend_class

    def remove_by_backend_id(self, backend_id):
        """
        Remove the backend class with the provided ``backend_id`` from the registry.
        """
        del self._backend_class_map[backend_id]

    def remove(self, backend_class):
        """
        Remove the provided backend class from the registry.
        """
        self.remove_by_backend_id(backend_class.get_backend_id())

    def __contains__(self, backend_class):
        """
        Returns ``True`` if the provided backend_class is in the registry.

        Parameters:
            backend_class: A subclass of :class:`.AbstractSmsBackend`.
        """
        return backend_class.get_backend_id() in self._backend_class_map

    def __iter__(self):
        """
        Returns an iterator over all backend classes in the registry.

        Same as :meth:`.iter_backend_classes`.
        """
        return self.iter_backend_classes()

    def iter_backend_classes(self):
        """
        Returns an iterator over all backend classes in the registry.
        """
        return iter(self._backend_class_map.values())

    def get_default_backend_id(self):
        """
        Get the default backend ID.

        Retrieved from the :setting:`IEVV_SMS_DEFAULT_BACKEND_ID` setting.

        Defaults to ``debugprint`` if the setting is not defined, or if it
        boolean False (None, empty string, ...).
        """
        return getattr(settings, 'IEVV_SMS_DEFAULT_BACKEND_ID', None) or 'debugprint'

    def get_backend_class_by_id(self, backend_id):
        """
        Get backend class by ID.

        Args:
            backend_id (str): The backend ID. If this is ``None``, we use
                the default backend (see :meth:`.get_default_backend_id`)
        """
        if backend_id is None:
            backend_id = self.get_default_backend_id()
        return self._backend_class_map[backend_id]

    def make_backend_instance(self, phone_number, message, backend_id=None, **kwargs):
        """
        Make a backend instance. Does not send the message.

        Args:
            phone_number (str): The phone number to send the message to.
            message (str): The message to send.
            backend_id (str): The ID of the backend to use for sending.
                If this is ``None``, we use the default backend
                (see :meth:`.get_default_backend_id`).
            **kwargs: Extra kwargs for the :class:`.AbstractSmsBackend`
                constructor.

        Returns:
            .AbstractSmsBackend: An instance of a subclass of :class:`.AbstractSmsBackend`.
        """
        backend_class = self.get_backend_class_by_id(backend_id=backend_id)
        backend = backend_class(phone_number=phone_number, message=message, **kwargs)
        return backend

    def send(self, phone_number, message, backend_id=None, **kwargs):
        """
        Send an SMS message.

        Shortcut for ``make_backend_instance(...).send()``.

        See :meth:`.make_backend_instance`.

        Args:
            phone_number: See :meth:`.make_backend_instance`.
            message: See :meth:`.make_backend_instance`.
            backend_id: See :meth:`.make_backend_instance`.
            **kwargs: See :meth:`.make_backend_instance`.

        Raises:
             django.core.exceptions.ValidationError: If validation of the phone number,
                message or kwargs fails.

        Returns:
            .AbstractSmsBackend: An instance of a subclass of :class:`.AbstractSmsBackend`.
        """
        backend = self.make_backend_instance(
            phone_number=phone_number, message=message,
            backend_id=backend_id, **kwargs)
        backend.clean()
        backend.send()
        return backend

    def get_backend_class(self, backend_id=None):
        if not backend_id:
            backend_id = self.get_default_backend_id()
        return self._backend_class_map[backend_id]


class MockableRegistry(Registry):
    """
    A non-singleton version of :class:`.Registry`. For tests.

    Typical usage in a test::

        from ievv_opensource.ievv_sms import sms_registry

        class MockSmsBackend(sms_registry.AbstractSmsBackend):
            def send(self):
                print('{}: {}'.format(self.clean_phone_number, self.cleaned_message))

        mockregistry = sms_registry.MockableRegistry()
        mockregistry.add(MockSmsBackend)

        with mock.patch('ievv_opensource.ievv_sms.sms_registry.Registry.get_instance',
                        lambda: mockregistry):
            pass  # ... your code here ...
    """

    def __init__(self):
        self._instance = None  # Ensure the singleton-check is not triggered
        super(MockableRegistry, self).__init__()


def send_sms(phone_number, message, backend_id=None, **kwargs):
    """
    Send SMS message.

    Just a shortcut for :meth:`.Registry.send`
    (``Registry.get_instance().send(...)``).

    Args:
        phone_number (str): The phone number to send the message to.
        message (str): The message to send.
        backend_id (str): The ID of the backend to use for sending.
            If this is ``None``, we use the default backend
            (see :meth:`.get_default_backend_id`).
        **kwargs: Extra kwargs for the :class:`.AbstractSmsBackend`
            constructor.

    Returns:
        .AbstractSmsBackend: An instance of a subclass of :class:`.AbstractSmsBackend`.
    """
    return Registry.get_instance().send(
        phone_number=phone_number,
        message=message,
        backend_id=backend_id,
        **kwargs)


def get_sms_part_count(message, backend_id=None):
    backend_class = Registry.get_instance().get_backend_class(backend_id=backend_id)
    return backend_class.get_part_count(message)


def get_sms_text_length(message, backend_id=None):
    backend_class = Registry.get_instance().get_backend_class(backend_id=backend_id)
    return backend_class.get_sms_text_length(message)


def get_sms_max_length(backend_id=None):
    backend_class = Registry.get_instance().get_backend_class(backend_id=backend_id)
    return backend_class.get_max_length()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2019-11-24 19:53
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('ievv_logging', '0003_auto_20191123_1147'),
    ]

    operations = [
        migrations.RenameField(
            model_name='ievvloggingeventbase',
            old_name='error_occured',
            new_name='success_run',
        ),
        migrations.RenameField(
            model_name='ievvloggingeventitem',
            old_name='error_occured',
            new_name='success_run',
        ),
    ]



from __future__ import unicode_literals
import os

from ievv_opensource.utils.ievvbuildstatic import cssbuildbaseplugin
from ievv_opensource.utils.ievvbuildstatic import filepath
from ievv_opensource.utils.ievvbuildstatic import utils
from ievv_opensource.utils.shellcommandmixin import ShellCommandError


class Plugin(cssbuildbaseplugin.AbstractPlugin):
    """
    SASS build plugin --- builds .scss files into css, and supports watching
    for changes.

    By default, we assume ``sassc`` is available on PATH, but you can
    override the path to the sassc executable by setting the
    ``IEVVTASKS_BUILDSTATIC_SASSC_EXECUTABLE`` environment variable.

    Examples:

        Very simple example where the source file is in
        ``demoapp/staticsources/styles/theme.scss``::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.sassbuild.Plugin(sourcefile='theme.scss'),
                    ]
                )
            )

        A more complex example that builds a django-cradmin theme
        where sources are split in multiple directories, and
        the bower install directory is on the scss path
        (the example also uses :class:`ievv_opensource.utils.ievvbuildstatic.bowerinstall.Plugin`)::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.bowerinstall.Plugin(
                            packages={
                                'bootstrap': '~3.1.1'
                            }
                        ),
                        ievvbuildstatic.sassbuild.Plugin(
                            sourcefolder='styles/cradmin_theme_demoapp',
                            sourcefile='theme.scss',
                            other_sourcefolders=[
                                'styles/cradmin_base',
                                'styles/cradmin_theme_default',
                            ],
                            sass_include_paths=[
                                'bower_components',
                            ]
                        )
                    ]
                )
            )
    """

    name = 'sassbuild'

    def __init__(self, sourcefile, sourcefolder='styles',
                 destinationfolder=None,
                 other_sourcefolders=None,
                 sass_include_paths=None,
                 sass_variables=None,
                 **kwargs):
        """
        Parameters:
            sourcefile: Main source file (the one including all other scss files)
                relative to ``sourcefolder``.
            sourcefolder: The folder where ``sourcefile`` is located relative to
                the source folder of the :class:`~ievv_opensource.utils.ievvbuild.config.App`.
                You can specify a folder in another app using
                a :class:`ievv_opensource.utils.ievvbuildstatic.filepath.SourcePath` object,
                but this is normally not recommended.
            sass_include_paths: Less include paths as a list. Paths are relative
                to the source folder of the :class:`~ievv_opensource.utils.ievvbuild.config.App`.
                You can specify folders in other apps using
                :class:`ievv_opensource.utils.ievvbuildstatic.filepath.SourcePath` objects
            **kwargs: Kwargs for :class:`ievv_opensource.utils.ievvbuildstatic.cssbuildbaseplugin.AbstractPlugin`.
        """
        self.sourcefolder = sourcefolder
        if isinstance(sourcefolder, filepath.FilePathInterface) and destinationfolder is None:
            raise ValueError('destinationfolder must be specifed when sourcefolder is not a string.')
        self.destinationfolder = destinationfolder or sourcefolder

        self.other_sourcefolders = other_sourcefolders or []
        self.sass_include_paths = sass_include_paths or []
        self.sourcefile = sourcefile
        self.sass_variables = sass_variables or {}
        super(Plugin, self).__init__(**kwargs)

    def install(self):
        super(Plugin, self).install()
        self.app.get_installer('npm').queue_install(
            'postcss-scss')

    def get_sourcefolder_path(self):
        return self.app.get_source_path(self.sourcefolder)

    def get_sourcefile_path(self):
        return self.app.get_source_path(self.sourcefolder, self.sourcefile)

    def get_destinationfile_path(self):
        return self.app.get_destination_path(
            self.destinationfolder, self.sourcefile, new_extension='.css')

    def get_other_sourcefolders_paths(self):
        return map(self.app.get_source_path, self.other_sourcefolders)

    def get_sass_include_paths(self, temporary_directory):
        sass_include_paths = list(map(self.app.get_source_path, self.sass_include_paths))
        sass_include_paths.append(self.get_importable_sass_directory(temporary_directory))
        return sass_include_paths

    def format_sass_include_paths(self, temporary_directory):
        if self.sass_include_paths:
            return ':'.join(self.get_sass_include_paths(temporary_directory))
        else:
            return ''

    def get_sassc_executable(self):
        return os.environ.get('IEVVTASKS_BUILDSTATIC_SASSC_EXECUTABLE', 'sassc')

    def get_sass_kwargs(self, temporary_directory):
        kwargs = {}
        sass_include_paths = self.format_sass_include_paths(temporary_directory=temporary_directory)
        if sass_include_paths:
            kwargs['load_path'] = sass_include_paths
        # style = 'nested'
        # if self.app.apps.is_in_production_mode():
        #     style = 'compressed'
        # kwargs['style'] = style
        return kwargs

    def get_importable_sass_directory(self, temporary_directory):
        importable_sass_directory_path = os.path.join(temporary_directory, 'automatically_on_sass_path')
        if not os.path.exists(importable_sass_directory_path):
            os.makedirs(importable_sass_directory_path)
        return importable_sass_directory_path

    def make_importable_sass_file(self, temporary_directory, sass_file_name, sass_code):
        directory_path = os.path.join(self.get_importable_sass_directory(temporary_directory),
                                      'autogenerated_partials')
        if not os.path.exists(directory_path):
            os.mkdir(directory_path)
        sass_file_path = os.path.join(directory_path, sass_file_name)
        with open(sass_file_path, 'wb') as f:
            f.write(sass_code.encode('utf-8'))

    def make_importable_variables_file_content(self):
        contentlist = []
        for variablename, value in self.sass_variables.items():
            contentlist.append('${}: {};'.format(variablename, value))
        header = '// Autogenerated by "ievv buildstatic". Do not edit this file.'
        return '{}\n\n{}'.format(header, '\n'.join(contentlist))

    def make_importable_variables_file(self, temporary_directory):
        self.make_importable_sass_file(temporary_directory=temporary_directory,
                                       sass_file_name='_variables.scss',
                                       sass_code=self.make_importable_variables_file_content())

    def preprocess_styles(self, temporary_directory):
        super(Plugin, self).preprocess_styles(temporary_directory)
        self.make_importable_variables_file(temporary_directory=temporary_directory)

    def build_css(self, temporary_directory):
        self.get_logger().command_start('Building {source} into {destination}.'.format(
            source=self.get_sourcefile_path(),
            destination=self.get_destinationfile_path()))

        destinationdirectory = os.path.dirname(self.get_destinationfile_path())
        if not os.path.exists(destinationdirectory):
            os.makedirs(destinationdirectory)

        executable = self.get_sassc_executable()
        try:
            self.run_shell_command(executable,
                                   args=[
                                       self.get_sourcefile_path(),
                                       self.get_destinationfile_path()
                                   ],
                                   kwargs=self.get_sass_kwargs(temporary_directory=temporary_directory))
        except ShellCommandError:
            self.get_logger().command_error('SASS build FAILED!')
            self.add_deferred_warning('SASS build FAILED!')
            raise cssbuildbaseplugin.CssBuildException()
        else:
            self.get_logger().command_success('SASS build successful :)')
            self.add_deferred_success('SASS build successful :)')

    def get_watch_folders(self):
        """
        We only watch the folder where the scss sources are located,
        so this returns the absolute path of the ``sourcefolder``.
        """
        folders = [self.app.get_source_path(self.sourcefolder)]
        if self.other_sourcefolders:
            folders.extend(self.get_other_sourcefolders_paths())
        return folders

    def get_watch_regexes(self):
        return ['^.+\.scss$']

    def __str__(self):
        return '{}({})'.format(super(Plugin, self).__str__(),
                               os.path.join(self.sourcefolder, self.sourcefile))

    def get_all_source_file_paths(self):
        regexfilelist = utils.RegexFileList(include_patterns=self.get_watch_regexes())
        source_file_paths = []
        sourcefolders = [self.get_sourcefolder_path()]
        sourcefolders.extend(self.get_other_sourcefolders_paths())
        for folder in sourcefolders:
            source_file_paths.extend(
                os.path.join(folder, relative_path)
                for relative_path in regexfilelist.get_files_as_list(folder))
        return source_file_paths



import sys

from ievv_opensource.utils.ievvdevrun.runnables import base


class RunnableThread(base.ShellCommandRunnableThread):
    """
    ``ievv buildstatic --watch`` runnable thread.

    Examples:

        You can just add it to your Django development settings with::

            IEVVTASKS_DEVRUN_RUNNABLES = {
                'default': ievvdevrun.config.RunnableThreadList(
                    ievvdevrun.runnables.ievv_buildstatic.RunnableThread()
                )
            }

    """
    default_autorestart_on_crash = True

    def get_logger_name(self):
        return 'ievv buildstatic --watch'

    def get_command_config(self):
        return {
            'executable': sys.executable,
            'args': ['manage.py', 'ievvtasks_buildstatic', '--watch']
        }



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import inspect
import os
import re
from collections import OrderedDict

import htmls
from django import template
from django.db import connection
from ievv_opensource.utils.singleton import Singleton


class AbstractCustomSql(object):
    """
    Defines custom SQL that can be executed by the ``ievv_customsql`` framework.

    You typically override :meth:`.initialize` and use :meth:`.execute_sql` to add
    triggers and functions, and override :meth:`.recreate_data` to rebuild the data
    maintained by the triggers, but many other use-cases are also possible.
    """
    def __init__(self, appname=None):
        """
        Args:
            appname: Not required - it is added automatically by :class:`.Registry`, and
                used by :meth:`.__str__`` for easier debugging / prettier output.
        """
        self.appname = appname

    def execute_sql(self, sql):
        """
        Execute the provided SQL.

        Args:
            sql (str): String of SQL.
        """
        cursor = connection.cursor()
        cursor.execute(sql)

    def execute_sql_multiple(self, sql_iterable):
        for sql in sql_iterable:
            self.execute_sql(sql)

    def make_sql_file_path(self, path):
        this_file_path = inspect.getfile(self.__class__)
        sqldirectory = os.path.join(
            os.path.dirname(inspect.getfile(self.__class__)),
            '{}_sqlcode'.format(os.path.splitext(this_file_path)[0]))
        return os.path.join(sqldirectory, path)

    def get_sql_from_file(self, path):
        """
        Get SQL from a file as a string.

        Args:
            path (str): A path relative to a directory named the same as the module
                where this class is located without the filename extension and suffixed
                with ``_sqlcode``.

                So if the file with this class is in ``path/to/customsql.py``,
                path will be relative to ``path/to/customsqlcode/``.
        """
        with open(self.make_sql_file_path(path), 'rb') as f:
            sql = f.read().decode('utf-8')
        return sql

    def execute_sql_from_file(self, path):
        """
        Execute SQL from the provided file.

        Args:
            path: See :meth:`.get_sql_from_file`.
        """
        self.execute_sql(self.get_sql_from_file(path))

    def execute_sql_from_files(self, paths):
        """
        Shortcut for calling :meth:`.execute_sql_from_file`
        with multiple files.

        Calls :meth:`.execute_sql_from_file` once for each file
        in ``paths``.

        Args:
            paths (list): A list of paths. See :meth:`.get_sql_from_file`
                for the format of each path.
        """
        for path in paths:
            self.execute_sql_from_file(path=path)

    def render_sql_from_template(self, path, context_data):
        template_sql = self.get_sql_from_file(path)
        context_data = context_data or {}
        djangotemplate = template.Template(template_sql)
        return djangotemplate.render(template.Context(context_data))

    def execute_sql_from_template_file(self, path, context_data=None):
        """
        Execute SQL from the provided Django template file.

        The SQL is retrieved from the file, then processed as a Django
        template with the provided ``context_data``, and the result
        is executed using :meth:`.execute_sql`.

        Args:
            path: See :meth:`.get_sql_from_file`.
            context_data (dict): Template context data. Can be ``None``.
        """
        self.execute_sql(self.render_sql_from_template(path=path, context_data=context_data))

    def execute_sql_from_template_files(self, paths, context_data=None):
        """
        Shortcut for calling :meth:`.execute_sql_from_template_file`
        with multiple files.

        Calls :meth:`.execute_sql_from_template_file` once for each file
        in ``paths``.

        Args:
            paths (list): A list of paths. See :meth:`.get_sql_from_file`
                for the format of each path.
            context_data (dict): Forwarded to :meth:`.execute_sql_from_template_file`.
        """
        for path in paths:
            self.execute_sql_from_template_file(path=path, context_data=context_data)

    def initialize(self):
        """
        Code to initialize the custom SQL.

        You should create triggers, functions, columns, indexes, etc. in this
        method, using :meth:`.execute_sql`, or using plain Django code.

        Make sure to write everything in a manner that updates or creates
        everything in a self-contained manner. This method is called both
        for the first initialization, and to update code after updates/changes.

        Must be overridden in subclasses.
        """
        raise NotImplementedError()

    def _normalize_whitespace(self, sql):
        return re.sub(r'\s+', ' ', sql).strip()

    CREATE_TRIGGER_PATTERN = re.compile(r'^\s*CREATE\s+TRIGGER\s+'
                                        r'(?P<trigger_name>[a-zA-Z0-9_]+)\s+'
                                        r'(?:BEFORE|AFTER|INSTEAD\s+OF)\s+'
                                        r'(?:[^\s]+\s+OR\s+)*(?:[^\s]+)\s+'
                                        r'ON\s+(?P<table_name>[a-zA-Z0-9_]+)',
                                        re.IGNORECASE | re.MULTILINE)

    def make_drop_trigger_statements_from_sql_code(self, sql):
        matches = self.CREATE_TRIGGER_PATTERN.finditer(sql)
        drop_statements = []
        for match in matches:
            drop_statement = 'DROP TRIGGER IF EXISTS {trigger_name} ON {table_name}'.format(
                **match.groupdict())
            drop_statements.append(drop_statement)
        return drop_statements

    CREATE_FUNCTION_PATTERN = re.compile(r'^\s*CREATE\s+(?:OR\s+REPLACE\s+)?FUNCTION\s+'
                                         r'(?P<function_name>[a-zA-Z0-9_]+)\s*'
                                         r'\((?P<arguments>[^)]*)\)',
                                         re.IGNORECASE | re.MULTILINE)

    def make_drop_function_statements_from_sql_code(self, sql):
        matches = self.CREATE_FUNCTION_PATTERN.finditer(sql)
        drop_statements = []
        for match in matches:
            function_name = match.groupdict()['function_name']
            arguments = self._normalize_whitespace(match.groupdict()['arguments'])
            drop_statement = 'DROP FUNCTION IF EXISTS {function_name} ({arguments}) RESTRICT'.format(
                function_name=function_name,
                arguments=arguments)
            drop_statements.append(drop_statement)
        return drop_statements

    CREATE_INDEX_PATTERN = re.compile(r'^\s*CREATE\s+(?:UNIQUE\s+)?INDEX\s+'
                                      r'(?:CONCURRENTLY\s+)?(?:IF\s+NOT\s+EXISTS\s+)?'
                                      r'(?P<index_name>[a-zA-Z0-9_]+)\s*ON',
                                      re.IGNORECASE | re.MULTILINE)

    def make_drop_index_statements_from_sql_code(self, sql):
        matches = self.CREATE_INDEX_PATTERN.finditer(sql)
        drop_statements = []
        for match in matches:
            drop_statement = 'DROP INDEX IF EXISTS {index_name} RESTRICT'.format(
                **match.groupdict())
            drop_statements.append(drop_statement)
        return drop_statements

    CREATE_TYPE_PATTERN = re.compile(r'^\s*CREATE\s+TYPE\s+'
                                     r'(?P<type_name>[a-zA-Z0-9_]+)',
                                     re.IGNORECASE | re.MULTILINE)

    def make_drop_type_statements_from_sql_code(self, sql):
        matches = self.CREATE_TYPE_PATTERN.finditer(sql)
        drop_statements = []
        for match in matches:
            drop_statement = 'DROP TYPE IF EXISTS {type_name} RESTRICT'.format(
                **match.groupdict())
            drop_statements.append(drop_statement)
        return drop_statements

    def make_drop_statements_from_sql_code(
            self, sql, sql_types=('index', 'type', 'function', 'trigger')):
        output_sql = []
        for sql_type in sql_types:
            methodname = 'make_drop_{}_statements_from_sql_code'.format(sql_type)
            output_sql.extend(getattr(self, methodname)(sql))
        return output_sql

    def make_drop_statements_from_sql_file(self, path):
        return self.make_drop_statements_from_sql_code(self.get_sql_from_file(path))

    def make_drop_statements_from_sql_files(self, paths):
        sql_statements = []
        for path in paths:
            sql_statements.extend(self.make_drop_statements_from_sql_file(path))
        return sql_statements

    def make_drop_statements_from_sql_template_file(self, path, context_data):
        sql = self.render_sql_from_template(path=path, context_data=context_data)
        return self.make_drop_statements_from_sql_code(sql)

    def make_drop_statements_from_sql_template_files(self, paths, context_data):
        sql_statements = []
        for path in paths:
            sql = self.render_sql_from_template(path=path, context_data=context_data)
            sql_statements.extend(self.make_drop_statements_from_sql_code(sql))
        return sql_statements

    def clear(self):
        """
        Revert :meth:`.initialize` and remove any data created by the triggers.

        Drop/delete triggers, functions, columns, indexes, etc. created
        in :meth:`.initialize`.
        """
        raise NotImplementedError()

    def recreate_data(self):
        """
        Recreate all data that any triggers created in :meth:`.initialize`
        would normally keep in sync automatically.

        Can not be used unless :meth:`.initialize` has already be run (at some point).
        This restriction is here to make it possible to create SQL functions
        in :meth:`.initialize` that this method uses to recreate the data. Without this
        restriction, code-reuse between :meth:`.initialize` and this function would be
        very difficult.
        """
        pass

    def run(self):
        """
        Run both :meth:`.initialize` and :meth:`.recreate_data`.
        """
        self.initialize()
        self.recreate_data()

    def __str__(self):
        return '{} in {}'.format(self.__class__.__name__, self.appname)


class Registry(Singleton):
    """
    Registry of :class:`.AbstractCustomSql` objects.

    Examples:

        First, define a subclass of :class:`.AbstractCustomSql`.

        Register the custom SQL class with the registry via an AppConfig for your
        Django app::

            from django.apps import AppConfig
            from ievv_opensource.ievv_customsql import customsql_registry
            from myapp import customsql

            class MyAppConfig(AppConfig):
                name = 'myapp'

                def ready(self):
                    customsql_registry.Registry.get_instance().add(customsql.MyCustomSql)

        See ``ievv_opensource/demo/customsql/apps.py`` for a complete demo.
    """

    def __init__(self):
        super(Registry, self).__init__()
        self._customsql_classes = []
        self._customsql_classes_by_appname_map = OrderedDict()

    def add(self, appname, customsql_class):
        """
        Add the given ``customsql_class`` to the registry.

        Parameters:
            appname: The django appname where the ``customsql_class`` belongs.
            customsql_class: A subclass of :class:`.AbstractCustomSql`.
        """
        if customsql_class in self._customsql_classes:
            raise ValueError('{}.{} is already in the custom SQL registry.'.format(
                customsql_class.__module__, customsql_class.__name__))
        self._customsql_classes.append(customsql_class)
        if appname not in self._customsql_classes_by_appname_map:
            self._customsql_classes_by_appname_map[appname] = []
        self._customsql_classes_by_appname_map[appname].append(customsql_class)

    def remove(self, appname, customsql_class):
        self._customsql_classes.remove(customsql_class)
        self._customsql_classes_by_appname_map[appname].remove(customsql_class)
        if len(self._customsql_classes_by_appname_map[appname]) == 0:
            del self._customsql_classes_by_appname_map[appname]

    def __contains__(self, customsql_class):
        """
        Returns ``True`` if the provided customsql_class is in the registry.

        Parameters:
            customsql_class: A subclass of :class:`.AbstractCustomSql`.
        """
        return customsql_class in self._customsql_classes

    def __iter__(self):
        """
        Iterate over all :class:`.AbstractCustomSql` subclasses registered
        in the registry. The yielded values are objects of the
        classes initialized with no arguments.
        """
        for appname in self.iter_appnames():
            for customsql in self.iter_customsql_in_app(appname):
                yield customsql

    def iter_appnames(self):
        """
        Returns an iterator over all the appnames in the registry.
        Each item in the iterator is an appname (a string).
        """
        return iter(self._customsql_classes_by_appname_map.keys())

    def iter_customsql_in_app(self, appname):
        """
        Iterate over all :class:`.AbstractCustomSql` subclasses registered
        in the provided appname. The yielded values are objects of the
        classes initialized with no arguments.
        """
        for customsql_class in self._customsql_classes_by_appname_map[appname]:
            yield customsql_class(appname)

    def run_all_in_app(self, appname):
        """
        Loops through all the :class:`.AbstractCustomSql` classes registered in the registry
        with the provided ``appname``, and call :meth:`.AbstractCustomSql.run` for each of them.
        """
        for customsql in self.iter_customsql_in_app(appname):
            customsql.run()

    def run_all(self):
        """
        Loops through all the :class:`.AbstractCustomSql` classes in the registry, and call
        :meth:`.AbstractCustomSql.run` for each of them.
        """
        for customsql in self:
            customsql.run()


class MockableRegistry(Registry):
    """
    A non-singleton version of :class:`.Registry`. For tests.

    Typical usage in a test::

        from ievv_opensource.ievv_customsql import customsql_registry

        class MockCustomSql(customsql_registry.AbstractCustomSql):
            # ...

        mockregistry = customsql_registry.MockableRegistry()
        mockregistry.add(MockCustomSql())

        with mock.patch('ievv_opensource.ievv_customsql.customsql_registry.Registry.get_instance',
                        lambda: mockregistry):
            pass  # ... your code here ...
    """

    def __init__(self):
        self._instance = None  # Ensure the singleton-check is not triggered
        super(MockableRegistry, self).__init__()



import arrow
import json
from psycopg2._range import DateTimeTZRange

from django.conf import settings
from rest_framework.fields import ModelField


class DateTimeRangeSerializerField(ModelField):
    """
    Add timezone to upper and lower datetime values if they do not contain TZ-formatting.

    We need to do this to store the correct UTC value in the database, so the stored
    value is not actually the local time.

    If the stored value is actually the local time stored as "UTC", we end
    being one hour behind if the timezone for Norway.
    """
    def to_internal_value(self, data):
        rel = self.model_field.remote_field
        require_tz_adjust = False
        if data:
            data_dict = json.loads(data)
            iso_lower = data_dict.get('lower')
            iso_upper = data_dict.get('upper')
            require_tz_adjust = (iso_lower and '+' not in iso_lower) or (iso_upper and '+' not in iso_upper)
        if rel is not None:
            return rel.model._meta.get_field(rel.field_name).to_python(data)
        result = self.model_field.to_python(data)
        if require_tz_adjust:
            lower = result.lower
            upper = result.upper
            if lower is not None:
                lower = arrow.get(result.lower, settings.TIME_ZONE).datetime
            if upper is not None:
                upper = arrow.get(result.upper, settings.TIME_ZONE).datetime
            return DateTimeTZRange(lower, upper)
        return result

    def __isoformat_datetime(self, datetime):
        if datetime:
            return arrow.get(datetime).to(settings.TIME_ZONE).format('YYYY-MM-DD HH:mm:ss')
        return None

    def to_representation(self, obj):
        datetimerange_value = getattr(obj, self.field_name, None)
        if not datetimerange_value:
            return None
        data = {
            "bounds": "[)",
            "lower": self.__isoformat_datetime(datetimerange_value.lower),
            "upper": self.__isoformat_datetime(datetimerange_value.upper),
        }
        return json.dumps(data)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2019-11-06 13:10
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ievv_logging', '0001_initial'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='ievvloggingeventbase',
            options={'ordering': ['-id']},
        ),
        migrations.AlterModelOptions(
            name='ievvloggingeventitem',
            options={'ordering': ['-id']},
        ),
        migrations.AlterField(
            model_name='ievvloggingeventitem',
            name='data',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='ievvloggingeventitem',
            name='time_spent_in_seconds',
            field=models.IntegerField(blank=True, null=True),
        ),
    ]



from __future__ import unicode_literals

import json
import os

import sh
from ievv_opensource.utils.ievvbuildstatic import pluginbase
from ievv_opensource.utils.ievvbuildstatic import utils
from ievv_opensource.utils.shellcommandmixin import ShellCommandMixin, ShellCommandError


class JavascriptMerger(object):
    def __init__(self):
        self.javascript_filepaths = []

    def add(self, javascript_filepath):
        self.javascript_filepaths.append(javascript_filepath)

    def __wrap_with_function_wrapper(self, source):
        sourcelines = []
        for line in source.splitlines():
            if line:
                line = '  {}'.format(line)
            sourcelines.append(line)

        sourcelines.insert(0, '(function() {')
        sourcelines.append('}).call(this);')
        return '\n'.join(sourcelines)

    def merge(self, rootdirectory, with_function_wrapper=True):
        output = []
        for javascript_filepath in self.javascript_filepaths:
            filecontent = open(javascript_filepath, 'rb').read().decode('utf-8')
            filecontent = filecontent.replace('\r\n', '\n').replace('\r', '\n')
            output.append('/* {} */'.format(
                            os.path.relpath(javascript_filepath, rootdirectory)))
            output.append(filecontent)
        source = '\n\n'.join(output)
        if with_function_wrapper:
            source = self.__wrap_with_function_wrapper(source=source)
        return source


class Plugin(pluginbase.Plugin, ShellCommandMixin):
    """
    CoffeeScript build plugin --- builds .coffee files into css, and supports watching
    for changes.

    Examples:

        Very simple example where the source file is in
        ``demoapp/staticsources/scripts/coffeescript/app.coffee``::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.coffeebuild.Plugin(destinationfile='app.js'),
                    ]
                )
            )

        Lets say we want build a different folder, and exclude all
        ``*.spec.coffee`` files in within the folder::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.coffeebuild.Plugin(
                            destinationfile='myapp.js',
                            sourcefolder=os.path.join('scripts', 'coffeescript', 'myapp')
                            sourcefile_exclude_patterns=['^.*\.spec\.coffee$']),
                    ]
                )
            )

        We also support linting, and it is enabled by default. You can disable
        linting with ``lint=False``, and you can override the coffeelint options.
        Lets say we want to change max line length to 120, and make breaking
        it a warning instead of an error::

            ievvbuildstatic.config.App(
                appname='demoapp',
                version='1.0.0',
                plugins=[
                    ievvbuildstatic.coffeebuild.Plugin(
                        destinationfile='app.js',
                        lintconfig={
                            "max_line_length": {
                                'value': 120,
                                'level': "warn",
                                'limitComments': True,
                            }
                        }
                    ),
                ]
            ),

    """

    name = 'coffeebuild'
    default_group = 'js'

    def __init__(self,
                 destinationfile,
                 sourcefile_include_patterns=None,
                 sourcefile_exclude_patterns=None,
                 sourcefolder=os.path.join('scripts', 'coffeescript'),
                 destinationfolder='scripts',
                 extra_watchfolders=None,
                 with_function_wrapper=True,
                 lint=True,
                 lintconfig=None,
                 **kwargs):
        """
        Parameters:
            destinationfile: Path to destination file relative to ``destinationfolder``.
            sourcefile_include_patterns: List of source file regexes. Same format as
                as for :class:`ievv_opensource.utils.ievvbuildstatic.utils.RegexFileList`.
                Defaults to ``['^.*\.coffee$']``.
            sourcefile_exclude_patterns: List of source file regexes. Same format as
                as for :class:`ievv_opensource.utils.ievvbuildstatic.utils.RegexFileList`.
            sourcefolder: The folder where sourcefiles is located relative to
                the source folder of the :class:`~ievv_opensource.utils.ievvbuild.config.App`.
                Defaults to ``scripts/coffeescript/``.
            destinationfolder: The folder where ``destinationfile`` is located relative to
                the destination folder of the :class:`~ievv_opensource.utils.ievvbuild.config.App`.
                Defaults to ``scripts/``.
            extra_watchfolders: List of extra folders to watch for changes.
                Relative to the source folder of the
                :class:`~ievv_opensource.utils.ievvbuild.config.App`.
            with_function_wrapper: Include function wrapper around the resulting
                javascript file.
            **kwargs: Kwargs for :class:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin`.
        """
        super(Plugin, self).__init__(**kwargs)
        self.destinationfile = destinationfile
        self.sourcefiles = utils.RegexFileList(
            include_patterns=sourcefile_include_patterns or ['^.*\.coffee$'],
            exclude_patterns=sourcefile_exclude_patterns
        )
        self.destinationfolder = destinationfolder
        self.sourcefolder = sourcefolder
        self.extra_watchfolders = extra_watchfolders or []
        self.with_function_wrapper = with_function_wrapper
        self.lint = lint
        self.lintconfig = lintconfig or {}

    def get_sourcefolder_path(self):
        return self.app.get_source_path(self.sourcefolder)

    def get_destinationfile_path(self):
        return self.app.get_destination_path(self.destinationfolder, self.destinationfile)

    def get_coffee_version(self):
        return None

    def get_coffeelint_version(self):
        return None

    def install(self):
        self.app.get_installer('npm').queue_install(
            'coffee-script', version=self.get_coffee_version())
        if self.lint:
            self.app.get_installer('npm').queue_install(
                'coffeelint', version=self.get_coffeelint_version())

    def get_coffee_executable(self):
        return self.app.get_installer('npm').find_executable('coffee')

    def get_coffeelint_executable(self):
        return self.app.get_installer('npm').find_executable('coffeelint')

    def get_all_sourcefiles(self):
        return self.sourcefiles.get_files_as_list(rootfolder=self.get_sourcefolder_path())

    def lint_coffeescript_file(self, relative_filepath, lintconfig_path):
        executable = self.get_coffeelint_executable()
        sourcefile_path = os.path.join(self.get_sourcefolder_path(), relative_filepath)
        self.get_logger().debug('CoffeeScript linting {!r}'.format(
            relative_filepath))
        try:
            self.run_shell_command(executable,
                                   args=[
                                       '-f', lintconfig_path,
                                       sourcefile_path
                                   ])
        except ShellCommandError:
            self.get_logger().error('CoffeeScript linting of {!r} FAILED!'.format(
                relative_filepath))

    def make_lintconfig_file(self, temporary_directory):
        lintconfig_path = os.path.join(temporary_directory, 'coffeelint.json')
        try:
            coffeelint = sh.Command(self.get_coffeelint_executable())
            default_lintconfig_json = coffeelint('--makeconfig',
                                                 _err=self.log_shell_command_stderr)
        except sh.ErrorReturnCode:
            self.get_logger().error('Failed to create coffeelint config file.')
            raise ShellCommandError()
        else:
            lintconfig = json.loads(str(default_lintconfig_json))
            lintconfig.update(self.lintconfig)
            open(lintconfig_path, 'wb').write(json.dumps(lintconfig, indent=2).encode('utf-8'))
            return lintconfig_path

    def lint_coffeescript_files(self, temporary_directory):
        lintconfig_path = self.make_lintconfig_file(temporary_directory=temporary_directory)
        for relative_filepath in self.get_all_sourcefiles():
            self.lint_coffeescript_file(relative_filepath=relative_filepath,
                                        lintconfig_path=lintconfig_path)

    def compile_coffescript_file(self, js_directory, relative_filepath):
        executable = self.get_coffee_executable()
        sourcefile_path = os.path.join(self.get_sourcefolder_path(), relative_filepath)
        destinationfile_directory = os.path.dirname(
            os.path.join(js_directory, relative_filepath))
        try:
            self.run_shell_command(executable,
                                   args=[
                                       '-b',
                                       '-o', destinationfile_directory,
                                       '-c', sourcefile_path,
                                   ])
        except ShellCommandError:
            self.get_logger().error('CoffeeScript build of {!r} FAILED!'.format(
                relative_filepath))
            raise
        else:
            self.get_logger().debug('CoffeeScript build of {!r} successful :)'.format(
                relative_filepath))

    def build_coffeescript_files(self, js_directory):
        for relative_filepath in self.get_all_sourcefiles():
            self.compile_coffescript_file(
                js_directory=js_directory,
                relative_filepath=relative_filepath)

    def merge_javascript_files(self, js_directory):
        javascript_merger = JavascriptMerger()
        for root, dirs, files in os.walk(js_directory):
            for filename in files:
                filepath = os.path.abspath(os.path.join(root, filename))
                javascript_merger.add(filepath)
        return javascript_merger.merge(rootdirectory=js_directory,
                                       with_function_wrapper=self.with_function_wrapper)

    def build(self, temporary_directory):
        if self.lint:
            self.lint_coffeescript_files(temporary_directory=temporary_directory)
        js_directory = os.path.join(temporary_directory, 'js')
        os.mkdir(js_directory)
        try:
            self.build_coffeescript_files(js_directory=js_directory)
        except ShellCommandError:
            self.get_logger().command_error('CoffeeScript build FAILED!')
        else:
            javascript_source = self.merge_javascript_files(js_directory=js_directory)
            destination_directory = os.path.dirname(self.get_destinationfile_path())
            if not os.path.exists(destination_directory):
                os.makedirs(destination_directory)
            open(self.get_destinationfile_path(), 'wb').write(
                javascript_source.encode('utf-8'))
            self.get_logger().command_success(
                'CoffeeScript build successful :) Output in {}'.format(
                    self.get_destinationfile_path()))

    def run(self):
        self.get_logger().command_start('CoffeeScript building all files in {!r} matching {}'.format(
            self.get_sourcefolder_path(),
            self.sourcefiles.prettyformat_patterns()))
        temporary_directory = self.make_temporary_build_directory()
        self.build(temporary_directory=temporary_directory)

    def get_extra_watchfolder_paths(self):
        return map(self.app.get_source_path, self.extra_watchfolders)

    def get_watch_folders(self):
        """
        We only watch the folder where the coffee sources are located,
        so this returns the absolute path of the ``sourcefolder``.
        """
        folders = [self.app.get_source_path(self.sourcefolder)]
        if self.extra_watchfolders:
            folders.extend(self.get_extra_watchfolder_paths())
        return folders

    def get_watch_regexes(self):
        return ['^.+[.]coffee$']

    def __str__(self):
        return '{}({})'.format(super(Plugin, self).__str__(), self.sourcefolder)



from ievv_opensource.utils.ievvbuildstatic import pluginbase
from ievv_opensource.utils.logmixin import Logger
from ievv_opensource.utils.shellcommandmixin import ShellCommandError
from ievv_opensource.utils.shellcommandmixin import ShellCommandMixin


class Plugin(pluginbase.Plugin, ShellCommandMixin):
    """
    Run a script from package.json (I.E.: ``npm run <something>``.

    Examples:

        Lets say you have the following in your package.json::

            {
              "scripts": {
                "myscript": "echo 'hello'"
              }
            }

        You can then make this script run with::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.npmrun.Plugin(script='myscript'),
                    ]
                )
            )
    """
    name = 'npmrun'

    def __init__(self, script, script_args=None, **kwargs):
        """

        Args:
            script (str): The name of the key for the script in the ``scripts``
                object in ``package.json``.
            script_args (list): Arguments for the script as a list of strings.
            **kwargs: Kwargs for :class:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin`.
        """
        self.script = script
        self.script_args = script_args
        super(Plugin, self).__init__(**kwargs)

    def run(self):
        name = '"{script}" package.json script'.format(script=self.script)
        self.get_logger().command_start('Running the {name} for {appname}'.format(
            name=name,
            appname=self.app.appname))
        try:
            self.app.get_installer('npm').run_packagejson_script(
                script=self.script,
                args=self.script_args or [])
        except ShellCommandError:
            self.get_logger().command_error('{name} FAILED!'.format(name=name))
            raise SystemExit()
        else:
            self.get_logger().command_success('{name} succeeded :)'.format(name=name))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.core.mail import send_mail
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = 'Send a test email.'

    def add_arguments(self, parser):
        parser.add_argument('--from', dest='from_email', default=settings.DEFAULT_FROM_EMAIL,
                            help='From email - defaults to DEFAULT_FROM_EMAIL setting.')
        parser.add_argument('--to', dest='to_email', required=True,
                            action='append',
                            help='To email address(es). Can be repeated to send to multiple emails.')
        parser.add_argument('--html', dest='html',
                            required=False, action='store_true', default=False,
                            help='Send html email.')
        parser.add_argument('--message-html', dest='message_html', default='',
                            help='Can be used with --html to override the HTML email message. If not '
                                 'specified, we use a lorem ipsum message.')
        parser.add_argument('--message-plain', dest='message_plain', default='',
                            help='Can be override the plain text email message. If not '
                                 'specified, we use a lorem ipsum message.')
        parser.add_argument('--use-cradmin-email', dest='use_cradmin_email',
                            required=False, action='store_true',
                            help='Use django_cradmin.apps.cradmin_email.emailutils.AbstractEmail '
                                 'instead of just the send_mail() function.')

    def get_plaintext_message(self):
        if self.options['message_plain']:
            return self.options['message_plain']
        return 'Maecenas faucibus mollis interdum. Cum sociis ' \
               'natoque penatibus et magnis dis parturient montes, ' \
               'nascetur ridiculus mus. Praesent commodo cursus magna, ' \
               'vel scelerisque nisl consectetur et. Fusce dapibus, ' \
               'tellus ac cursus commodo, tortor mauris condimentum ' \
               'nibh, ut fermentum massa justo sit amet risus. Etiam ' \
               'porta sem malesuada magna mollis euismod. Cras mattis ' \
               'consectetur purus sit amet fermentum. Donec ullamcorper ' \
               'nulla non metus auctor fringilla.' \
               '\n\n' \
               'Duis mollis, est non commodo luctus, nisi erat porttitor ' \
               'ligula, eget lacinia odio sem nec elit. Maecenas sed diam ' \
               'eget risus varius blandit sit amet non magna. Vivamus ' \
               'sagittis lacus vel augue laoreet rutrum faucibus dolor ' \
               'auctor. Maecenas sed diam eget risus varius blandit sit ' \
               'amet non magna. Curabitur blandit tempus porttitor. ' \
               'Etiam porta sem malesuada magna mollis euismod. Vestibulum ' \
               'id ligula porta felis euismod semper.'

    def get_html_message(self):
        if self.options['message_html']:
            return self.options['message_html']
        return 'Maecenas <strong>faucibus</strong> mollis interdum. Cum sociis ' \
               'natoque penatibus et magnis dis parturient montes, ' \
               'nascetur ridiculus mus. Praesent commodo cursus magna, ' \
               'vel scelerisque nisl consectetur et. Fusce dapibus, ' \
               'tellus ac cursus commodo, tortor mauris condimentum ' \
               'nibh, ut fermentum massa justo sit amet risus. Etiam ' \
               'porta sem malesuada magna mollis euismod. Cras mattis ' \
               'consectetur purus sit amet fermentum. Donec ullamcorper ' \
               'nulla non metus auctor fringilla.' \
               '<br><br>' \
               '<em>Duis mollis</em>, est non commodo luctus, nisi erat porttitor ' \
               'ligula, eget lacinia odio sem nec elit. Maecenas sed diam ' \
               'eget risus varius blandit sit amet non magna. Vivamus ' \
               'sagittis lacus vel augue laoreet rutrum faucibus dolor ' \
               'auctor. <a href="http://example.com">Maecenas sed</a> ' \
               'diam eget risus varius blandit sit ' \
               'amet non magna. Curabitur blandit tempus porttitor. ' \
               'Etiam porta sem malesuada magna mollis euismod. Vestibulum ' \
               'id ligula porta felis euismod semper.'

    def get_email_kwargs(self):
        kwargs = {
            'subject': 'Test email',
            'message': self.get_plaintext_message(),
            'from_email': self.options['from_email'],
            'recipient_list': self.options['to_email']
        }
        if self.options['html']:
            kwargs['html_message'] = self.get_html_message()
        return kwargs

    def make_cradmin_email(self):
        from django_cradmin.apps.cradmin_email import emailutils

        class CradminEmail(emailutils.AbstractEmail):
            html_message_template = 'ievv_developemail/cradmin_email/html_message.django.html'

            def __init__(self, subject, html_message=None, **kwargs):
                self.subject = subject
                extra_context_data = kwargs.pop('extra_context_data', None) or {}
                extra_context_data['html_message'] = html_message
                kwargs['extra_context_data'] = extra_context_data
                super().__init__(**kwargs)

            def render_subject(self):
                return self.subject

        email_kwargs = self.get_email_kwargs()
        return CradminEmail(subject=email_kwargs['subject'],
                            from_email=email_kwargs['from_email'],
                            recipient_list=email_kwargs['recipient_list'],
                            html_message=email_kwargs['html_message'])

    def handle(self, *args, **options):
        self.options = options
        use_cradmin_email = options['use_cradmin_email']
        if use_cradmin_email and not options['html']:
            self.stderr.write('--use-cradmin-email requires --html')
            raise SystemExit()
        if use_cradmin_email:
            self.make_cradmin_email().send()
        else:
            send_mail(**self.get_email_kwargs())



from django.utils import timezone


class ProgressPrintIterator:
    """
    Progress print iterator. Useful to print progress of long running scripts.

    Example::

        queryset = MyModel.objects
        total = queryset.count()
        for obj, is_end_of_group in ProgressPrintIterator(
                iterator=queryset.iterator(),
                total_count=total,
                what='Doing something',
                items_per_group=500):
            # Do something with ``obj``. If you want to do something after 500 items has been processed
            # including the last iteration (which may be less than 500 items),
            # use ``if is_end_of_group``
    """
    def __init__(self, iterator, total_count, what, items_per_group=500, log_function=None):
        """

        Args:
            iterator: Some iterator, such as a ``queryset.iterator()``.
            total_count: Total number of items.
            what: A message to print when printing progress
            items_per_group: Items per group - we print progress each time we have processed this number of items.
            log_function: A log function. For management scripts, you want to set this to ``self.stdout.write``.
        """
        self.iterator = iterator
        self.total_count = total_count
        self.what = what
        self.items_per_group = items_per_group
        self.log_function = log_function or print

    def __iter__(self):
        start_time = timezone.now()
        for index, item in enumerate(self.iterator, start=1):
            progress_percent = index / self.total_count * 100
            is_end_of_group = (index % self.items_per_group == 0) or (index == self.total_count)
            yield item, is_end_of_group
            if is_end_of_group:
                now = timezone.now()
                time_used = now - start_time
                if progress_percent > 0:
                    estimated_end_delta = time_used / progress_percent * (100 - progress_percent)
                    estimated_end_minutes = round(estimated_end_delta.total_seconds() / 60, 2)
                else:
                    estimated_end_minutes = 'UNKNOWN'
                self.log_function(
                    f'{round(progress_percent, 1)}% [{index}/{self.total_count}]: {self.what}. '
                    f'Est. minutes remaining: {estimated_end_minutes}')



from django.conf import settings
from django.core.management.base import BaseCommand
from ievv_opensource.utils.logmixin import Logger


class Command(BaseCommand):
    help = 'A build system (for sass, javascript, ...).'

    def add_arguments(self, parser):
        parser.add_argument('-w', '--watch', dest='watch',
                            required=False, action='store_true',
                            help='Starts a blocking process that watches for changes.')
        parser.add_argument('-a', '--appname', dest='appnames',
                            required=False, default=[], action='append',
                            help='Specify one or more apps to build. If not specified, all '
                                 'apps are built. Example: "-a ievv_jsbase -a demoapp2"')
        parser.add_argument('-s', '--skip', dest='skipgroups',
                            required=False, action='append',
                            help='Skip a plugin group. Can be repeated. Example: "-s test -s css".')
        parser.add_argument('--skip-js', dest='skipgroups',
                            required=False, action='append_const', const='js',
                            help='Skip building javascript. The same as "--skip js".')
        parser.add_argument('--skip-css', dest='skipgroups',
                            required=False, action='append_const', const='css',
                            help='Skip building css. The same as "--skip css".')
        parser.add_argument('--skip-jstests', dest='skipgroups',
                            required=False, action='append_const', const='jstest',
                            help='Skip running javascript tests. The same as "--skip jstest".')
        parser.add_argument('--skip-slow-tests', dest='skipgroups',
                            required=False, action='append_const', const='slow-jstest',
                            help='Skip running slow javascript tests. The same as "--skip slow-jstest". '
                                 'Slow javascript tests are also skipped if you use "--skip-jstests".')
        parser.add_argument('-i', '--include', dest='includegroups',
                            required=False, action='append',
                            help='Include plugin groups. If this is provided, only plugin '
                                 'belonging to the provided groups are executed. '
                                 'Overrides any groups provided using "--skip" (I.E.: included '
                                 'groups are included no matter what other options are provided).')
        parser.add_argument('--loglevel', dest='loglevel',
                            required=False, default='stdout',
                            choices=['debug', 'stdout', 'info', 'warning', 'error'],
                            help='Set loglevel. Can be one of: debug, stdout, info, warning or error '
                                 '(listed in order of verbosity).')
        parser.add_argument('-q', '--quiet', dest='loglevel',
                            required=False, action='store_const', const='warning',
                            help='Quiet output. Same as "--loglevel warning".')
        parser.add_argument('--production', dest='production_mode',
                            required=False, action='store_true', default=False,
                            help='Build in production mode.')
        parser.add_argument('--debug', dest='loglevel',
                            required=False, action='store_const', const='debug',
                            help='Verbose output. Same as "--loglevel debug".')
        settings.IEVVTASKS_BUILDSTATIC_APPS.add_cli_arguments(parser=parser)

    def __get_skipgroups(self, options):
        skipgroups = options['skipgroups']
        skipgroups = set(skipgroups or [])
        if 'skip-jstests' in skipgroups:
            skipgroups.add('skip-slow-jstests')
        if 'skip-js' in skipgroups:
            skipgroups.add('skip-slow-jstests')
            skipgroups.add('skip-jstests')
        return skipgroups

    def __get_includegroups(self, options):
        includegroups = options['includegroups']
        includegroups = set(includegroups or [])
        if 'skip-jstests' in includegroups:
            includegroups.add('skip-slow-jstests')
        if 'skip-js' in includegroups:
            includegroups.add('skip-slow-jstests')
            includegroups.add('skip-jstests')
        return includegroups

    def handle(self, *args, **options):
        loglevel = getattr(Logger, options['loglevel'].upper())
        appnames = options['appnames']
        production_mode = options['production_mode']
        skipgroups = self.__get_skipgroups(options=options)
        includegroups = self.__get_includegroups(options=options)
        if appnames:
            try:
                settings.IEVVTASKS_BUILDSTATIC_APPS.validate_appnames(appnames=appnames)
            except ValueError as error:
                self.stderr.write(str(error))
                raise SystemExit()
        settings.IEVVTASKS_BUILDSTATIC_APPS.configure_logging(
            loglevel=loglevel,
            command_error_message='Re-run with "--debug" for more details.')
        if production_mode:
            settings.IEVVTASKS_BUILDSTATIC_APPS.set_production_mode()
        else:
            settings.IEVVTASKS_BUILDSTATIC_APPS.set_development_mode()
        settings.IEVVTASKS_BUILDSTATIC_APPS.set_options(options=options)
        settings.IEVVTASKS_BUILDSTATIC_APPS.log_help_header()
        settings.IEVVTASKS_BUILDSTATIC_APPS.install(appnames=appnames,
                                                    skipgroups=skipgroups,
                                                    includegroups=includegroups)
        settings.IEVVTASKS_BUILDSTATIC_APPS.run(appnames=appnames,
                                                skipgroups=skipgroups,
                                                includegroups=includegroups)
        if options['watch']:
            settings.IEVVTASKS_BUILDSTATIC_APPS.watch(appnames=appnames,
                                                      skipgroups=skipgroups,
                                                      includegroups=includegroups)



import time

import sys
import signal


class RunnableThreadList(object):
    """
    List of :class:`~ievv_opensource.utils.ievvdevrun.runnables.base.AbstractRunnableThread`
    objects.

    You use this with the :setting:`IEVVTASKS_DEVRUN_RUNNABLES` setting
    to define what to run with ``ievv devrun``.
    """
    def __init__(self, *runnablethreads):
        """
        Parameters:
            runnablethreads: :class:`~ievv_opensource.utils.ievvdevrun.runnables.base.AbstractRunnableThread`
                objects to add to the list.
        """
        self._runnablethreads = []
        for runnablethread in runnablethreads:
            self.append(runnablethread)

    def append(self, runnablethread):
        """
        Append a :class:`~ievv_opensource.utils.ievvdevrun.runnables.base.AbstractRunnableThread`.

        Parameters:
            runnablethread: A :class:`~ievv_opensource.utils.ievvdevrun.runnables.base.AbstractRunnableThread`
                object.
        """
        self._runnablethreads.append(runnablethread)

    def start(self):
        """
        Start all the runnable threads and block until ``SIGTERM`` or
        ``KeyboardInterrupt``.
        """
        def start_all():
            for runnablethread in self._runnablethreads:
                runnablethread.start()

        def stop_all():
            for runnablethread in reversed(self._runnablethreads):
                runnablethread.stop()

        def join_all():
            for runnablethread in self._runnablethreads:
                runnablethread.join()

        def on_terminate(*args):
            print('')
            print('*' * 70)
            print('ievvdevrun terminated')
            print('*' * 70)
            stop_all()
            join_all()
            sys.exit(0)

        start_all()
        signal.signal(signal.SIGTERM, on_terminate)
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            stop_all()
        join_all()



import os
import shutil

from ievv_opensource.utils.ievvbuildstatic import filepath
from ievv_opensource.utils.ievvbuildstatic import pluginbase


class Plugin(pluginbase.Plugin):
    """
    Media copy plugin --- copies media files from staticsources into
    the static directory, and supports watching for file changes.

    Examples:

        Copy all files in ``demoapp/staticsources/demoapp/media/`` into
        ``demoapp/static/1.0.0/media/``::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.mediacopy.Plugin(),
                    ]
                )
            )

        Using a different source folder. This will copy
        all files in ``demoapp/staticsources/demoapp/assets/`` into
        ``demoapp/static/1.0.0/assets/``::

            IEVVTASKS_BUILDSTATIC_APPS = ievvbuildstatic.config.Apps(
                ievvbuildstatic.config.App(
                    appname='demoapp',
                    version='1.0.0',
                    plugins=[
                        ievvbuildstatic.mediacopy.Plugin(sourcefolder="assets"),
                    ]
                )
            )
    """
    name = 'mediacopy'

    def __init__(self, sourcefolder='media', destinationfolder=None, **kwargs):
        """
        Parameters:
            sourcefolder: The folder where media files is located relative to
                the source folder of the :class:`~ievv_opensource.utils.ievvbuild.config.App`.
                You can specify a folder in another app using
                a :class:`ievv_opensource.utils.ievvbuildstatic.filepath.SourcePath` object.
            **kwargs: Kwargs for :class:`ievv_opensource.utils.ievvbuildstatic.pluginbase.Plugin`.
        """
        super(Plugin, self).__init__(**kwargs)
        self.sourcefolder = sourcefolder
        if isinstance(sourcefolder, filepath.FilePathInterface) and destinationfolder is None:
            raise ValueError('destinationfolder must be specifed when sourcefolder is not a string.')
        self.destinationfolder = destinationfolder or sourcefolder

    def get_sourcefolder_path(self):
        return self.app.get_source_path(self.sourcefolder)

    def get_destinationfolder_path(self):
        return self.app.get_destination_path(self.destinationfolder)

    def run(self):
        sourcefolder = self.get_sourcefolder_path()
        self.get_logger().command_start('Copying media')
        if not os.path.exists(sourcefolder):
            self.get_logger().warning('Media source folder, {}, does not exist.'.format(
                sourcefolder))
            return

        destinationfolder = self.get_destinationfolder_path()
        if os.path.abspath(destinationfolder) == os.path.abspath(sourcefolder):
            raise ValueError('mediacopy: Sourcefolder {} is the same as destinationfolder {}'.format(
                sourcefolder, destinationfolder))

        if os.path.exists(destinationfolder):
            self.get_logger().debug('Removing {}'.format(destinationfolder))
            shutil.rmtree(destinationfolder)

        self.get_logger().debug('Copying {} -> {}'.format(sourcefolder, destinationfolder))
        shutil.copytree(sourcefolder, destinationfolder)
        self.get_logger().command_success('Copied media successfully :)')

    def get_watch_folders(self):
        """
        We only watch the folder where the less sources are located,
        so this returns the absolute path of the ``sourcefolder``.
        """
        return [self.app.get_source_path(self.sourcefolder)]

    def __str__(self):
        return '{}({})'.format(super(Plugin, self).__str__(), self.sourcefolder)



from django.core.exceptions import ValidationError
from ievv_opensource.utils import choices_with_meta


class ChoiceSetupError(Exception):
    """
    Raised by :meth:`.SerializableChoicesWithMeta.validate_setup`.
    """


class AbstractChoiceConstants:
    """
    Base class used to define constants for some set of choices.

    You should define the choice values as attributes of the subclass,
    and then override :meth:`.get_choice_kwargs` to provide
    values for the :meth:`.make_choice` method.
    """
    @classmethod
    def get_default_choice_class(cls):
        raise NotImplementedError()

    @classmethod
    def get_choice_kwargs(cls):
        """
        Returns a dict mapping choice values to kwargs for the
        :class:`ievv_opensource.utils.choices_with_meta.Choice`
        corresponding to the value.
        """
        return {}

    @classmethod
    def make_choice(cls, value, choice_class=None, **kwargs):
        choice_class = choice_class or cls.get_default_choice_class()
        full_kwargs = {}
        for kwarg_key, kwarg_value in cls.get_choice_kwargs()[value].items():
            full_kwargs.setdefault(kwarg_key, kwarg_value)
        full_kwargs.update(kwargs)
        full_kwargs['value'] = value
        return choice_class(**full_kwargs)


class SerializableChoice(choices_with_meta.Choice):
    """
    Base class for choices within a :class:`.SerializableChoicesWithMeta`.
    """
    def validate_setup(self, choiceswithmeta_object, **kwargs):
        """
        Validate that this choice has sane attributes within
        the context of the provided ``choiceswithmeta_object``.

        See :meth:`.SerializableChoicesWithMeta.validate_setup` for more
        details.

        Args:
            choiceswithmeta_object (.SerializableChoicesWithMeta): The
                choices with meta object that contains this choice.
        """


class ImpliesAllExcept:
    def __init__(self, *except_choice_values):
        self.except_choice_values = except_choice_values

    def __call__(self, choiceswithmeta_object, source_choice):
        implies = set()
        for choice in choiceswithmeta_object.iterchoices():
            if choice.value not in self.except_choice_values:
                implies.add(choice.value)
        return implies


class ImpliesAll(ImpliesAllExcept):
    def __init__(self):
        super().__init__()


class SerializableMultiChoice(SerializableChoice):
    """
    Base class for choices within a :class:`.SerializableMultiChoicesWithMeta`.
    """
    def __init__(self, *args, implies=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.raw_implies = implies or set()
        self._implies = None
        self._implied_by = None

        # Set in :class:`.ChoicesWithMeta.add`.
        self._choices_with_meta = None

    def _build_implies(self):
        if not self._choices_with_meta:
            raise ValueError('Can not use the "implies" property before the choice has '
                             'been added to a MultipleChoicesWithMeta.')
        if callable(self.raw_implies):
            raw_implies_callable = self.raw_implies
            implies = raw_implies_callable(choiceswithmeta_object=self._choices_with_meta,
                                           source_choice=self)
        else:
            implies = set(self.raw_implies)
        all_implies = set()
        for permission in implies:
            if permission == self.value:
                continue
            try:
                implied_choice = self._choices_with_meta[permission]
            except KeyError:
                raise ChoiceSetupError(
                    '{}: {!r} implies {!r}, but {!r} does '
                    'not exist.'.format(self.classpath, self.value, permission, permission))
            all_implies.add(implied_choice.value)
            all_implies.update(implied_choice.implies)
        return list(all_implies)

    @property
    def implies(self):
        if self._implies is None:
            self._implies = self._build_implies()
        return self._implies

    def _build_implied_by(self):
        if not self._choices_with_meta:
            raise ValueError('Can not use the "implied_by" property before the choice has '
                             'been added to a MultipleChoicesWithMeta.')
        implied_by = set()
        for choice in self._choices_with_meta.iterchoices():
            if self.value in choice.implies:
                implied_by.add(choice.value)
        return list(implied_by)

    @property
    def implied_by(self):
        if self._implied_by is None:
            self._implied_by = self._build_implied_by()
        return self._implied_by

    @property
    def value_and_implied_by(self):
        values = {self.value}
        values.update(self.implied_by)
        return list(values)

    def as_serializable_data(self):
        data = super().as_serializable_data()
        data['implies'] = list(sorted(self.implies))
        data['implied_by'] = list(sorted(self.implied_by))
        return data


class SerializableChoicesWithMeta(choices_with_meta.ChoicesWithMeta):
    """
    Extends :class:`ievv_opensource.utils.choices_with_meta.ChoicesWithMeta` with
    support for serializing choices.
    """
    @property
    def classpath(self):
        return '{}.{}'.format(self.__class__.__module__, self.__class__.__name__)

    def iterchoices_for_as_serializable_data(self):
        """
        This is here so that subclasses can filter the
        choices exposed via the APIs. This is called by
        :meth:`.as_serializable_data` to loop the choices
        to serialize, and those serialized choices are
        exposed in the APIs.
        """
        return self.iterchoices()

    def as_serializable_data(self):
        """
        Get as a serializable list of choices.
        """
        result = []
        for choice in self.iterchoices_for_as_serializable_data():
            result.append(choice.as_serializable_data())
        return result

    def validate_choice(self, choice):
        if not isinstance(choice, SerializableChoice):
            raise TypeError(
                f'Can only add objects of the '
                f'{SerializableChoice.get_classpath()} type '
                f'(or a subclass) to a SerializableChoicesWithMeta.')

    def add(self, choice):
        self.validate_choice(choice=choice)
        super().add(choice)

    def get_model_attribute_name(self):
        raise NotImplementedError()

    def clean_choice(self, choice, model_object):
        raise NotImplementedError()

    def clean_model_object(self, model_object):
        from django.utils.translation import ugettext
        value = getattr(model_object, self.get_model_attribute_name())
        if value not in self:
            raise ValidationError({
                self.get_model_attribute_name(): ugettext('"%(value)s" is not one of the available choices') % {
                    'value': value
                }
            })
        choice = self[value]
        self.clean_choice(choice, model_object)

    def validate_setup(self, **kwargs):
        """
        Validate that this ChoicesWithMeta contains sane choices.

        Can both raise exceptions (typically :exc:`.ChoiceSetupError`) and
        log warnings depending on the severity of the problems discovered.

        This is typically called from the ready() method of an AppConfig.

        By default, this just loops over all the choices and
        call :meth:`.SerializableChoice.validate_setup`.
        """
        for choice in self.iterchoices():
            choice.validate_setup(choiceswithmeta_object=self, **kwargs)


class SerializableMultiChoicesWithMeta(SerializableChoicesWithMeta):
    """
    ChoicesWithMeta class with special handling of multichoice fields.

    The choices added must be :class:`.SerializableMultiChoice` objects.
    """
    def normalize_values(self, values):
        """
        Normalize the provided ``values``, removing any values that is implied by another value.
        """
        values = set(values)
        implies = set()
        for permission in values:
            implies.update(self[permission].implies)
        return values - implies

    def validate_choice(self, choice):
        if not isinstance(choice, SerializableMultiChoice):
            raise TypeError(
                f'Can only add objects of the '
                f'{SerializableMultiChoice.get_classpath()} type '
                f'(or a subclass) to a SerializableMultiChoicesWithMeta.')
        if choice._choices_with_meta is not None:
            raise ValueError('Can not add the same Choice object to multiple ChoicesWithMeta.')

    def add(self, choice):
        super().add(choice)
        choice._choices_with_meta = self

    def clean_model_object(self, model_object):
        from django.utils.translation import ugettext
        for value in getattr(model_object, self.get_model_attribute_name()):
            if value not in self:
                raise ValidationError({
                    self.get_model_attribute_name(): ugettext('"%(value)s" is not one of the available choices') % {
                        'value': value
                    }
                })
        setattr(
            model_object, self.get_model_attribute_name(),
            self.normalize_values(getattr(model_object, self.get_model_attribute_name()))
        )
        for value in getattr(model_object, self.get_model_attribute_name()):
            choice = self[value]
            self.clean_choice(choice, model_object)

        # Maintain sort order to keep the values clean in the database and APIs
        values = list(getattr(model_object, self.get_model_attribute_name()))
        values.sort()
        setattr(model_object, self.get_model_attribute_name(), values)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.9 on 2019-10-31 11:08
from __future__ import unicode_literals

import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='IevvLoggingEventBase',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('slug', models.CharField(db_index=True, max_length=255, unique=True)),
                ('last_started', models.DateTimeField()),
                ('last_finished', models.DateTimeField(blank=True, null=True)),
                ('time_spent', models.CharField(blank=True, default='', max_length=255)),
                ('time_spent_in_seconds', models.IntegerField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='IevvLoggingEventItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('time_spent', models.CharField(blank=True, default='', max_length=255)),
                ('time_spent_in_seconds', models.IntegerField()),
                ('data', django.contrib.postgres.fields.jsonb.JSONField()),
                ('start_datetime', models.DateTimeField()),
                ('end_datetime', models.DateTimeField(blank=True, null=True)),
                ('created_datetime', models.DateTimeField(auto_now_add=True)),
                ('logging_base', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='ievv_logging.IevvLoggingEventBase')),
            ],
        ),
    ]



import os
import shutil

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError
from ievv_opensource.ievvtasks_common.open_file import open_file_with_default_os_opener
from ievv_opensource.utils import virtualenvutils
from ievv_opensource.utils.logmixin import Logger, LogMixin
from ievv_opensource.utils.shellcommandmixin import ShellCommandMixin, ShellCommandError


class Command(BaseCommand, LogMixin, ShellCommandMixin):
    help = 'Build, open and clean docs (or all at once).'

    def __get_documentation_directory(self):
        default_directory = os.path.join('not_for_deploy', 'docs')
        return getattr(settings, 'IEVVTASKS_DOCS_DIRECTORY', default_directory)

    def __get_documentation_build_directory(self):
        default_build_directory = os.path.join(self.__get_documentation_directory(), '_build')
        return getattr(settings, 'IEVVTASKS_DOCS_BUILD_DIRECTORY', default_build_directory)

    def __get_documentation_indexhtml_path(self):
        return os.path.join(self.__get_documentation_build_directory(), 'index.html')

    def add_arguments(self, parser):
        parser.add_argument('-b', '--build', dest='build_docs',
                            required=False, action='store_true',
                            help='Build the docs.')
        parser.add_argument('-c', '--clean', dest='cleandocs',
                            required=False, action='store_true',
                            help='Remove any existing built docs before building the docs.')
        parser.add_argument('-o', '--open', dest='opendocs',
                            required=False, action='store_true',
                            help='Open the docs after building them.')
        parser.add_argument('--fail-on-warning', dest='fail_on_warning',
                            required=False, action='store_true',
                            help='Fail on warning.')
        parser.add_argument('--buildstatic-docs', dest='include_buildstatic_docs',
                            action='store_false', default=False,
                            help='Include ievv buildstatic docs.')
        parser.add_argument('--loglevel', dest='loglevel',
                            required=False, default='stdout',
                            choices=['debug', 'stdout', 'info', 'warning', 'error'],
                            help='Set loglevel. Can be one of: debug, stdout, info, warning or error '
                                 '(listed in order of verbosity).')
        parser.add_argument('-q', '--quiet', dest='loglevel',
                            required=False, action='store_const', const='warning',
                            help='Quiet output. Same as "--loglevel warning".')
        parser.add_argument('--debug', dest='loglevel',
                            required=False, action='store_const', const='debug',
                            help='Verbose output. Same as "--loglevel debug".')

    def _opendocs(self):
        self.get_logger().info('Opening {} in your browser.'.format(self.indexhtmlpath))
        open_file_with_default_os_opener(self.indexhtmlpath)

    def __clean_docs(self):
        if os.path.exists(self.__get_documentation_build_directory()):
            self.get_logger().debug('Removing {}'.format(self.__get_documentation_build_directory()))
            shutil.rmtree(self.__get_documentation_build_directory())
        else:  # pragma: no cover
            # Excluded from code coverage because this does not happen
            # since we always run tests with a removed build directory.
            self.get_logger().debug('Not removing {} - it does not exist.'.format(
                self.__get_documentation_build_directory()))

    def log_shell_command_stderr(self, line):
        """
        Called by :meth:`.run_shell_command` each time the shell
        command outputs anything to stderr.
        """
        method = 'stderr'
        if 'warning' in line.lower():
            method = 'warning'
        elif 'error' in line.lower():
            method = 'error'
        getattr(self.get_logger(), method)(line.rstrip())

    def get_logger_name(self):
        return 'ievv docs'

    def get_loglevel(self):
        return self.loglevel

    def __build_sphinx_docs(self):
        kwargs = {
            'b': 'html'
        }
        if self.fail_on_warning:  # pragma: no cover
            # Excluded from coverage because testing this would
            # require a complete mock of the docs directory.
            kwargs['W'] = True
            kwargs['T'] = True
        try:
            self.run_shell_command(
                executable='sphinx-build',
                args=[
                    self.__get_documentation_directory(),
                    self.__get_documentation_build_directory(),
                ],
                kwargs=kwargs
            )
        except ShellCommandError:  # pragma: no cover
            # We do not need to show any more errors here - they
            # have already been printed by the _out and _err handlers.
            # Excluded from code coverage because we do not test fail_on_warning
            # as explained above.
            pass

    def __get_buildstatic_documentation_build_directory(self):
        return os.path.join(self.__get_documentation_build_directory(), 'ievvbuildstatic')

    def __build_buildstatic_docs(self):
        settings.IEVVTASKS_BUILDSTATIC_APPS.configure_logging(
            loglevel=self.get_loglevel(),
            command_error_message='Re-run with "--debug" for more details.')
        settings.IEVVTASKS_BUILDSTATIC_APPS.install()
        settings.IEVVTASKS_BUILDSTATIC_APPS.run()
        settings.IEVVTASKS_BUILDSTATIC_APPS.build_docs(
            output_directory=self.__get_buildstatic_documentation_build_directory())

    def handle(self, *args, **options):
        self.loglevel = getattr(Logger, options['loglevel'].upper())
        cleandocs = options['cleandocs']
        build_docs = options['build_docs']
        opendocs = options['opendocs']
        include_buildstatic_docs = options['include_buildstatic_docs']
        self.indexhtmlpath = self.__get_documentation_indexhtml_path()
        self.fail_on_warning = options['fail_on_warning']

        if not (cleandocs or opendocs or build_docs):
            raise CommandError('You must specify at least one of: --build, '
                               '--clean or --open. See --help for more info.')

        if cleandocs and opendocs and not build_docs:
            raise CommandError('Can not use --clean and --open without also using --build.')

        virtualenvutils.add_virtualenv_bin_directory_to_path()
        if cleandocs:
            self.__clean_docs()
        if build_docs:
            self.__build_sphinx_docs()
            if include_buildstatic_docs:
                self.__build_buildstatic_docs()
        if opendocs:
            self._opendocs()
        else:
            self.get_logger().info('Built docs. Use "ievv docs -o" to open them in your browser to view them.')
